# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Handle local file: links.
"""

import re
import os
import urllib.parse
import urllib.request
from datetime import datetime, timezone

from . import urlbase, get_index_html
from .. import log, LOG_CHECK, fileutil, mimeutil, LinkCheckerError, url as urlutil
from ..bookmarks import firefox
from .const import (WARN_FILE_MISSING_SLASH, WARN_FILE_SYSTEM_PATH,
                    WARN_FILE_ANCHORCHECK_DIRECTORY)


def get_files(dirname):
    """Get iterator of entries in directory. Only allows regular files
    and directories, no symlinks."""
    for entry in os.listdir(dirname):
        fullentry = os.path.join(dirname, entry)
        if os.path.islink(fullentry):
            continue
        if os.path.isfile(fullentry):
            yield entry
        elif os.path.isdir(fullentry):
            yield entry + "/"


def prepare_urlpath_for_nt(path):
    """
    URLs like *file://server/path/* result in a path named */server/path*.
    However urllib.url2pathname expects *////server/path*.
    """
    if '|' not in path:
        return "////" + path.lstrip("/")
    return path


def get_nt_filename(path):
    """Return case sensitive filename for NT path."""
    unc, rest = os.path.splitdrive(path)
    head, tail = os.path.split(rest)
    if not tail:
        return path
    for fname in os.listdir(unc + head):
        if fname.lower() == tail.lower():
            return os.path.join(get_nt_filename(unc + head), fname)
    log.error(LOG_CHECK, "could not find %r in %r", tail, head)
    return path


def get_os_filename(path):
    """Return filesystem path for given URL path."""
    if os.name == 'nt':
        path = prepare_urlpath_for_nt(path)
    res = urllib.request.url2pathname(fileutil.path_safe(path))
    if os.name == 'nt' and res.endswith(':') and len(res) == 2:
        # Work around https://bugs.python.org/issue11474
        res += os.sep
    return res


def is_absolute_path(path):
    """Check if given path is absolute. On Windows absolute paths start
    with a drive letter. On all other systems absolute paths start with
    a slash."""
    if os.name == 'nt':
        if re.search(r"^[a-zA-Z]:", path):
            return True
        path = path.replace("\\", "/")
    return path.startswith("/")


class FileUrl(urlbase.UrlBase):
    """
    Url link with file scheme.
    """

    def init(
        self,
        base_ref,
        base_url,
        parent_url,
        recursion_level,
        aggregate,
        line,
        column,
        page,
        name,
        url_encoding,
        extern,
    ):
        """Initialize the scheme."""
        super().init(
            base_ref,
            base_url,
            parent_url,
            recursion_level,
            aggregate,
            line,
            column,
            page,
            name,
            url_encoding,
            extern,
        )
        self.scheme = 'file'

    def build_base_url(self):
        """The URL is normed according to the platform:
         - the base URL is made an absolute *file://* URL
         - under Windows platform the drive specifier is normed
        """
        if self.base_url is None:
            return
        base_url = self.base_url
        if not (self.parent_url or self.base_ref or base_url.startswith("file:")):
            base_url = os.path.expanduser(base_url)
            if not is_absolute_path(base_url):
                try:
                    base_url = os.getcwd() + "/" + base_url
                except OSError as msg:
                    # occurs on stale remote filesystems (eg. NFS)
                    errmsg = _(
                        "Could not get current working directory: %(msg)s"
                    ) % dict(msg=msg)
                    raise LinkCheckerError(errmsg)
                if os.path.isdir(base_url):
                    base_url += "/"
            base_url = "file://" + base_url
        if os.name == "nt":
            base_url = base_url.replace("\\", "/")
            # transform c:/windows into /c|/windows
            base_url = re.sub("^file://(/?)([a-zA-Z]):", r"file:///\2|", base_url)
            # transform file://path into file:///path
            base_url = re.sub("^file://([^/])", r"file:///\1", base_url)
        self.base_url = base_url

    def build_url(self):
        """
        Calls super.build_url() and adds a trailing slash to directories.
        """
        self.build_base_url()
        if self.parent_url is not None:
            # URL joining with the parent URL only works if the query
            # of the base URL are removed first.
            # Otherwise the join function thinks the query is part of
            # the file name.
            urlparts = list(urllib.parse.urlsplit(self.base_url))
            # ignore query part for filesystem urls
            urlparts[3] = ''
            self.base_url = urlutil.urlunsplit(urlparts)
        super().build_url()
        # ignore query and fragment url parts for filesystem urls
        self.urlparts[3] = self.urlparts[4] = ''
        if self.is_directory() and not self.urlparts[2].endswith('/'):
            self.add_warning(
                _("Added trailing slash to directory."), tag=WARN_FILE_MISSING_SLASH
            )
            self.urlparts[2] += '/'
        self.url = urlutil.urlunsplit(self.urlparts)

    def add_size_info(self):
        """Get size of file content and modification time from filename path."""
        if self.is_directory():
            # Directory size always differs from the customer index.html
            # that is generated. So return without calculating any size.
            return
        filename = self.get_os_filename()
        self.size = fileutil.get_size(filename)
        self.modified = datetime.fromtimestamp(
            fileutil.get_mtime(filename), tz=timezone.utc)

    def check_connection(self):
        """
        Try to open the local file. Under NT systems the case sensitivity
        is checked.
        """
        if self.parent_url is not None and not self.parent_url.startswith("file:"):
            msg = _(
                "local files are only checked without parent URL or when"
                " the parent URL is also a file"
            )
            raise LinkCheckerError(msg)
        if self.is_directory():
            self.set_result(_("directory"))
        else:
            url = fileutil.path_safe(self.url)
            self.url_connection = urllib.request.urlopen(url)
            self.check_case_sensitivity()

    def check_case_sensitivity(self):
        """
        Check if url and windows path name match cases
        else there might be problems when copying such
        files on web servers that are case sensitive.
        """
        if os.name != 'nt':
            return
        path = self.get_os_filename()
        realpath = get_nt_filename(path)
        if path != realpath:
            self.add_warning(
                _(
                    "The URL path %(path)r is not the same as the"
                    " system path %(realpath)r. You should always use"
                    " the system path in URLs."
                )
                % {"path": path, "realpath": realpath},
                tag=WARN_FILE_SYSTEM_PATH,
            )

    def read_content(self):
        """Return file content, or in case of directories a dummy HTML file
        with links to the files."""
        if self.is_directory():
            data = get_index_html(get_files(self.get_os_filename()))
        else:
            data = super().read_content()
        return data

    def get_os_filename(self):
        """
        Construct os specific file path out of the *file://* URL.

        @return: file name
        @rtype: string
        """
        return get_os_filename(self.urlparts[2])

    def get_temp_filename(self):
        """Get filename for content to parse."""
        return self.get_os_filename()

    def is_directory(self):
        """
        Check if file is a directory.

        @return: True iff file is a directory
        @rtype: bool
        """
        filename = self.get_os_filename()
        return os.path.isdir(filename) and not os.path.islink(filename)

    def is_parseable(self):
        """Check if content is parseable for recursion.

        @return: True if content is parseable
        @rtype: bool
        """
        if self.is_directory():
            return True
        if firefox.has_sqlite and firefox.extension.search(self.url):
            return True
        return self.is_content_type_parseable()

    def set_content_type(self):
        """Set URL content type, or an empty string if content
        type could not be found."""
        if self.url:
            self.content_type = mimeutil.guess_mimetype(self.url, read=self.get_content)
        else:
            self.content_type = ""
        log.debug(LOG_CHECK, "MIME type: %s", self.content_type)

    def get_intern_pattern(self, url=None):
        """Get pattern for intern URL matching.

        @return non-empty regex pattern or None
        @rtype String or None
        """
        if url is None:
            url = self.url
        if not url:
            return None
        if url.startswith('file://'):
            i = url.rindex('/')
            if i > 6:
                # remove last filename to make directory internal
                url = url[: i + 1]
        return re.escape(url)

    def add_url(self, url, line=0, column=0, page=0, name="", base=None):
        """If a local webroot directory is configured, replace absolute URLs
        with it. After that queue the URL data for checking."""
        webroot = self.aggregate.config["localwebroot"]
        if webroot and url and url.startswith("/"):
            url = webroot + url[1:]
            log.debug(LOG_CHECK, "Applied local webroot `%s' to `%s'.", webroot, url)
        super().add_url(url, line=line, column=column, page=page, name=name, base=base)


class AnchorCheckFileUrl(FileUrl):
    """
    File URL link for AnchorCheck plugin.
    """

    def reset(self):
        super().reset()
        # the local file URI
        self.url_without_anchor = None

    def build_url(self):
        """
        Calls UrlBase.build_url() and adds a trailing slash to directories.
        """
        self.build_base_url()
        if self.parent_url is not None:
            # URL joining with the parent URL only works if the query
            # of the base URL are removed first.
            # Otherwise the join function thinks the query is part of
            # the file name.
            urlparts = list(urllib.parse.urlsplit(self.base_url))
            # ignore query part for filesystem urls
            urlparts[3] = ''
            self.base_url = urlutil.urlunsplit(urlparts)
        super(FileUrl, self).build_url()
        # ignore query url part for filesystem urls
        self.urlparts[3] = ''
        if self.is_directory() and not self.urlparts[2].endswith('/'):
            self.add_warning(
                _("Added trailing slash to directory."), tag=WARN_FILE_MISSING_SLASH
            )
            self.urlparts[2] += '/'
        self.url = urlutil.urlunsplit(self.urlparts)
        self.url_without_anchor = urlutil.urlunsplit(self.urlparts[:4] + [''])

    def check_connection(self):
        """
        Try to open the local file. Under NT systems the case sensitivity
        is checked.
        """
        if self.parent_url is not None and not self.parent_url.startswith("file:"):
            msg = _(
                "local files are only checked without parent URL or when"
                " the parent URL is also a file"
            )
            raise LinkCheckerError(msg)
        if self.is_directory():
            if self.anchor:
                self.add_warning(
                    _(
                        "URL `%s' is a directory with an anchor."
                        " When checking local files AnchorCheck does not support"
                        " anchors for directories."
                    ) % self.url,
                    tag=WARN_FILE_ANCHORCHECK_DIRECTORY,
                )

            self.set_result(_("directory"))
        else:
            url = fileutil.path_safe(self.url_without_anchor)
            self.url_connection = urllib.request.urlopen(url)
            self.check_case_sensitivity()

    def set_content_type(self):
        """Set URL content type, or an empty string if content
        type could not be found."""
        if self.url:
            self.content_type = mimeutil.guess_mimetype(
                self.url_without_anchor, read=self.get_content)
        else:
            self.content_type = ""
        log.debug(LOG_CHECK, "MIME type: %s", self.content_type)



# Copyright (C) 2004-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Special container classes.
"""


class LFUCache(dict):
    """Limited cache which purges least frequently used items."""

    def __init__(self, size=1000):
        """Initialize internal LFU cache."""
        super().__init__()
        if size < 1:
            raise ValueError("invalid cache size %d" % size)
        self.size = size

    def __setitem__(self, key, val):
        """Store given key/value."""
        if key in self:
            # store value, do not increase number of uses
            super().__getitem__(key)[1] = val
        else:
            super().__setitem__(key, [0, val])
            # check for size limit
            if len(self) > self.size:
                self.shrink()

    def shrink(self):
        """Shrink ca. 5% of entries."""
        trim = int(0.05 * len(self))
        if trim:
            items = super().items()
            # sorting function for items
            def keyfunc(x): return x[1][0]
            values = sorted(items, key=keyfunc)
            for item in values[0:trim]:
                del self[item[0]]

    def __getitem__(self, key):
        """Update key usage and return value."""
        entry = super().__getitem__(key)
        entry[0] += 1
        return entry[1]

    def uses(self, key):
        """Get number of uses for given key (without increasing the number of
        uses)"""
        return super().__getitem__(key)[0]

    def get(self, key, def_val=None):
        """Update key usage if found and return value, else return default."""
        if key in self:
            return self[key]
        return def_val

    def setdefault(self, key, def_val=None):
        """Update key usage if found and return value, else set and return
        default."""
        if key in self:
            return self[key]
        self[key] = def_val
        return def_val

    def items(self):
        """Return list of items, not updating usage count."""
        return [(key, value[1]) for key, value in super().items()]

    def iteritems(self):
        """Return iterator of items, not updating usage count."""
        for key, value in super().items():
            yield (key, value[1])

    def values(self):
        """Return list of values, not updating usage count."""
        return [value[1] for value in super().values()]

    def itervalues(self):
        """Return iterator of values, not updating usage count."""
        for value in super().values():
            yield value[1]

    def popitem(self):
        """Remove and return an item."""
        key, value = super().popitem()
        return (key, value[1])

    def pop(self):
        """Remove and return a value."""
        value = super().pop()
        return value[1]



__app_name__ = "LinkChecker"
__version__ = "10.6.0"
__release_date__ = "2025-07-28"
__copyright_year__ = "2025"
__author__ = "LinkChecker Authors"
__url__ = "https://linkchecker.github.io/linkchecker/"
__support_url__ = "https://github.com/linkchecker/linkchecker/issues"



# Copyright (C) 2010-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Define http test support classes for LinkChecker tests.
"""
import os
import time
import threading
from ftplib import FTP

from tests import skip
from . import LinkCheckTest


TIMEOUT = 5


class FtpServerTest(LinkCheckTest):
    """Start/stop an FTP server that can be used for testing."""

    def __init__(self, methodName="runTest"):
        """Init test class and store default ftp server port."""
        super().__init__(methodName=methodName)
        self.host = "localhost"
        self.port = None

    def setUp(self):
        """Start a new FTP server in a new thread."""
        super().setUp()
        self.port = start_server(self.host, 0)
        self.assertFalse(self.port is None)

    def tearDown(self):
        """Send stop request to server."""
        try:
            stop_server(self.host, self.port)
        except Exception:
            pass


def start_server(host, port):
    def line_logger(self, msg):
        if "kill" in msg:
            raise KeyboardInterrupt()

    try:
        from pyftpdlib.authorizers import DummyAuthorizer
        from pyftpdlib.handlers import FTPHandler
        from pyftpdlib.servers import FTPServer
        from pyftpdlib import __ver__ as pyftpdlib_version
    except ImportError:
        skip("pyftpdlib is not available")
        return
    authorizer = DummyAuthorizer()
    datadir = os.path.join(os.path.dirname(__file__), "data")
    authorizer.add_anonymous(datadir)

    # Instantiate FTP handler class
    ftp_handler = FTPHandler
    ftp_handler.authorizer = authorizer
    ftp_handler.timeout = TIMEOUT
    ftp_handler.logline = line_logger

    # Define a customized banner (string returned when client connects)
    ftp_handler.banner = "pyftpdlib %s based ftpd ready." % pyftpdlib_version

    # Instantiate FTP server class and listen to host:port
    address = (host, port)
    server = FTPServer(address, ftp_handler)
    port = server.address[1]
    t = threading.Thread(None, server.serve_forever)
    t.start()
    # wait for server to start up
    tries = 0
    while tries < 5:
        tries += 1
        try:
            ftp = FTP()
            ftp.connect(host, port, TIMEOUT)
            ftp.login()
            ftp.close()
            break
        except Exception:
            time.sleep(0.5)
    return port


def stop_server(host, port):
    """Stop a running FTP server."""
    ftp = FTP()
    ftp.connect(host, port, TIMEOUT)
    ftp.login()
    try:
        ftp.sendcmd("kill")
    except EOFError:
        pass
    ftp.close()



# Copyright (C) 2004-2009 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test miscellaneous html tag parsing and URL types
"""
import sys

from tests import need_network
from . import LinkCheckTest


class TestMisc(LinkCheckTest):
    """
    Test misc link types.
    """

    @need_network
    def test_misc(self):
        self.file_test("misc.html")

    def test_html5(self):
        self.file_test("html5.html")

    def test_utf8(self):
        self.file_test("utf8.html")

    @need_network
    def test_archive(self):
        self.file_test("archive.html")

    @need_network
    def test_itms_services(self):
        if sys.version_info < (3, 12):
            url = "itms-services:" \
                  "?action=download-manifest&url=https://linkchecker.github.io/"
        else:
            url = "itms-services://" \
                  "?action=download-manifest&url=https://linkchecker.github.io/"
        resultlines = [
            "url %s" % url,
            "cache key %s" % url,
            "real url %s" % url,
            "valid",
            "url https://linkchecker.github.io/",
            "cache key https://linkchecker.github.io/",
            "real url https://linkchecker.github.io/",
            "valid",
        ]
        self.direct(url, resultlines, recursionlevel=1)



import sys
import re
import csv
import requests

iana_uri_schemes = "https://www.iana.org/assignments/uri-schemes/uri-schemes.xhtml"
# CSV format: URI Scheme,Template,Description,Status,Well-Known URI Support,Reference,Notes
csv_iana_uri_schemes = (
    "https://www.iana.org/assignments/uri-schemes/uri-schemes-1.csv"
)

iana_uri_schemes_dict = {}
iana_uri_schemes_other = {
    "clsid": "Microsoft specific",
    "find": "Mozilla specific",
    "gemini": "Gemini protocol",
    "isbn": "ISBN (int. book numbers)",
    "javascript": "JavaScript",
    "litecoin": "Litecoin payment",
    "ltclightning": "Litecoin Lightning invoice",
    "ms-windows-store": "Microsoft Store",
    "slack": "Slack Technologies client",
    "tg": "Telegram",
    "whatsapp": "WhatsApp",
}

filter_uri_schemes_permanent = (
    "file",
    "ftp",
    "http",
    "https",
    "mailto",
)

template = '''
# from %(uri)s
ignored_schemes_permanent = r"""
%(permanent)s
"""

ignored_schemes_provisional = r"""
%(provisional)s
"""

ignored_schemes_historical = r"""
%(historical)s
"""

ignored_schemes_other = r"""
%(other)s
"""

ignored_schemes = "^({}{}{}{})$".format(
    ignored_schemes_permanent,
    ignored_schemes_provisional,
    ignored_schemes_historical,
    ignored_schemes_other,
)
ignored_schemes_re = re.compile(ignored_schemes, re.VERBOSE)

is_unknown_scheme = ignored_schemes_re.match
'''


def main(args):
    parse_csv_file(csv_iana_uri_schemes, iana_uri_schemes_dict)
    for scheme in iana_uri_schemes_other:
        if (
            scheme in iana_uri_schemes_dict["Permanent"]
            or scheme in iana_uri_schemes_dict["Provisional"]
            or scheme in iana_uri_schemes_dict["Historical"]
        ):
            raise ValueError(scheme)
    for scheme in filter_uri_schemes_permanent:
        if scheme in iana_uri_schemes_dict["Permanent"]:
            del iana_uri_schemes_dict["Permanent"][scheme]
    args = dict(
        uri=iana_uri_schemes,
        permanent=get_regex(iana_uri_schemes_dict["Permanent"]),
        provisional=get_regex(iana_uri_schemes_dict["Provisional"]),
        historical=get_regex(iana_uri_schemes_dict["Historical"]),
        other=get_regex(iana_uri_schemes_other),
    )
    res = template % args
    print(res.rstrip())
    return 0


def get_regex(schemes):
    expr = [
        f"|{re.escape(scheme).ljust(10)} # {description}"
        for scheme, description in sorted(schemes.items())
    ]
    return "\n".join(expr)


def parse_csv_file(url, res):
    """Parse given URL and write res with {scheme -> description}"""
    response = requests.get(url, stream=True)
    reader = csv.reader(response.iter_lines(decode_unicode=True))
    first_row = True
    for row in reader:
        if first_row:
            # skip first row
            first_row = False
        else:
            scheme, template, description, status, urisupport, reference, notes = row
            scheme = scheme.replace(" (OBSOLETE)", "")  # remove the HTTP historic experiments flag
            if status not in res:
                res[status] = {}
            res[status][scheme] = description


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python
# Copyright (C) 2011 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Example to filter XML output.
Call with XML output filename as first argument.
Prints filtered result on standard output.
"""
import sys
from xml.etree.ElementTree import parse


def main(args):
    filename = args[0]
    with open(filename) as fd:
        tree = parse(fd)
        filter_tree(tree)
        tree.write(sys.stdout, encoding="utf-8")


def filter_tree(tree):
    """Filter all 401 errors."""
    to_remove = []
    for elem in tree.findall("urldata"):
        valid = elem.find("valid")
        if (
            valid is not None
            and valid.text == "0"
            and valid.attrib.get("result", "").startswith("401")
        ):
            to_remove.append(elem)
    root = tree.getroot()
    for elem in to_remove:
        root.remove(elem)


if __name__ == "__main__":
    main(sys.argv[1:])



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Support for managing threads.
"""
import threading


class StoppableThread(threading.Thread):
    """Thread class with a stop() method. The thread itself has to check
    regularly for the stopped() condition."""

    def __init__(self):
        """Store stop event."""
        super().__init__()
        self._stopper = threading.Event()

    def stop(self):
        """Set stop event."""
        self._stopper.set()

    def stopped(self, timeout=None):
        """Return True if stop event is set."""
        return self._stopper.wait(timeout)



# Copyright (C) 2004-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test https.
"""
import datetime
from unittest.mock import patch

from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from OpenSSL import crypto

from .httpserver import HttpsServerTest, CookieRedirectHttpRequestHandler
from .. import get_file

from linkcheck import httputil


class TestHttps(HttpsServerTest):
    """
    Test https: link checking.
    """

    def __init__(self, methodName="runTest"):
        super().__init__(methodName=methodName)
        self.handler = CookieRedirectHttpRequestHandler

    @classmethod
    def setUpClass(cls):
        key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
        )

        with open(get_file("https_key.pem"), "wb") as f:
            f.write(key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),
            ))

        subject = issuer = x509.Name([
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, "LinkChecker"),
            x509.NameAttribute(NameOID.COMMON_NAME, "linkchecker.github.io"),
        ])

        cert = x509.CertificateBuilder().subject_name(
            subject
        ).issuer_name(
            issuer
        ).public_key(
            key.public_key()
        ).serial_number(
            x509.random_serial_number()
        ).not_valid_before(
            datetime.datetime.now(datetime.timezone.utc)
        ).not_valid_after(
            datetime.datetime(2119, 1, 2, 3, 4, 5)
        ).add_extension(
            x509.SubjectAlternativeName([x509.DNSName("localhost")]),
            critical=False,
        ).sign(key, hashes.SHA256())

        with open(get_file("https_cert.pem"), "wb") as f:
            f.write(cert.public_bytes(serialization.Encoding.PEM))

    def test_https(self):
        url = self.get_url("")
        resultlines = [
            "url %s" % url,
            "cache key %s" % url,
            "real url %s" % url,
            "valid",
        ]
        confargs = dict(sslverify=get_file("https_cert.pem"))
        with patch.dict("os.environ",
                        {"REQUESTS_CA_BUNDLE": get_file("https_cert.pem")}):
            self.direct(url, resultlines, recursionlevel=0, confargs=confargs)

    def test_x509_to_dict(self):
        with open(get_file("https_cert.pem"), "rb") as f:
            cert = crypto.load_certificate(crypto.FILETYPE_PEM, f.read())
        self.assertEqual(
            httputil.x509_to_dict(cert)["notAfter"], "Jan 02 03:04:05 2119 GMT"
        )



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
HTML parser test function.
"""


def pretty_print_html(fd, soup):
    """
    Print out all parsed HTML data,
    writing to the given file descriptor.

    @param fd: file like object
    @type fd: file
    @param soup: BeautifulSoup object
    @type soup: BeautifulSoup
    """
    for element in soup.find_all(True):
        tag = element.name
        element_text = element.text.strip()

        fd.write("<%s" % tag.replace("/", ""))
        for key, val in sorted(element.attrs.items()):
            if val is None:
                fd.write(" %s" % key)
            else:
                fd.write(f' {key}="{quote_attrval(val)}"')
        if element_text:
            fd.write(f">{element_text}</{tag}>")
        else:
            fd.write("/>")


def quote_attrval(s):
    """
    Quote a HTML attribute to be able to wrap it in double quotes.

    @param s: the attribute string to quote
    @type s: string
    @return: the quoted HTML attribute
    @rtype: string
    """
    res = []
    for c in s:
        if ord(c) <= 127:
            # ASCII
            if c == "&":
                res.append("&amp;")
            elif c == '"':
                res.append("&quot;")
            else:
                res.append(c)
        else:
            res.append("&#%d;" % ord(c))
    return "".join(res)



# Copyright (C) 2005-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

def x509_to_dict(x509):
    """Parse a x509 pyopenssl object to a dictionary with keys
    subject, subjectAltName and optional notAfter.
    """
    from cryptography.x509 import DNSName, SubjectAlternativeName

    crypto_cert = x509.to_cryptography()
    ext = crypto_cert.extensions.get_extension_for_class(SubjectAlternativeName)

    res = {
        'subject': ((('commonName', x509.get_subject().CN),),),
        'subjectAltName': [
            ('DNS', value) for value in ext.value.get_values_for_type(DNSName)]
    }
    try:
        # cryptography >= 42.0.0
        notAfter = crypto_cert.not_valid_after_utc
    except AttributeError:
        notAfter = crypto_cert.not_valid_after
    if notAfter is not None:
        res['notAfter'] = notAfter.strftime('%b %d %H:%M:%S %Y GMT')
    return res


def get_content_type(headers):
    """
    Get the MIME type from the Content-Type header value, or
    'application/octet-stream' if not found.

    @return: MIME type
    @rtype: string
    """
    ptype = headers.get('Content-Type', 'application/octet-stream')
    if ";" in ptype:
        # split off not needed extension info
        ptype = ptype.split(';')[0]
    return ptype.strip().lower()



# Copyright (C) 2004-2012 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
FTP checking.
"""
from .. import need_pyftpdlib
from .ftpserver import FtpServerTest


class TestFtp(FtpServerTest):
    """Test ftp: link checking."""

    @need_pyftpdlib
    def test_ftp(self):
        # ftp two slashes
        url = "ftp://%s:%d/" % (self.host, self.port)
        resultlines = [
            "url %s" % url,
            "cache key %s" % url,
            "real url %s" % url,
            "valid",
        ]
        self.direct(url, resultlines)
        # ftp use/password
        user = "anonymous"
        passwd = "Ftp"
        url = "ftp://%s:%s@%s:%d/" % (user, passwd, self.host, self.port)
        resultlines = [
            "url %s" % url,
            "cache key %s" % url,
            "real url %s" % url,
            "valid",
        ]
        self.direct(url, resultlines)
        # ftp one slash
        url = "ftp:/%s:%d/" % (self.host, self.port)
        nurl = self.norm(url)
        resultlines = [
            "url %s" % url,
            "cache key None",
            "real url %s" % nurl,
            "error",
        ]
        self.direct(url, resultlines)
        # missing path
        url = "ftp://%s:%d" % (self.host, self.port)
        nurl = self.norm(url)
        resultlines = [
            "url %s" % url,
            "cache key %s" % nurl,
            "real url %s" % nurl,
            "valid",
        ]
        self.direct(url, resultlines)
        # missing trailing dir slash
        url = "ftp://%s:%d/base" % (self.host, self.port)
        nurl = self.norm(url)
        resultlines = [
            "url %s" % url,
            "cache key %s" % nurl,
            "real url %s/" % nurl,
            "warning Missing trailing directory slash in ftp url.",
            "valid",
        ]
        self.direct(url, resultlines)
        # ftp two dir slashes
        url = "ftp://%s:%d//base/" % (self.host, self.port)
        nurl = self.norm(url)
        resultlines = [
            "url %s" % url,
            "cache key %s" % nurl,
            "real url %s" % nurl,
            "valid",
        ]
        self.direct(url, resultlines)
        # ftp many dir slashes
        url = "ftp://%s:%d////////base/" % (self.host, self.port)
        nurl = self.norm(url)
        resultlines = [
            "url %s" % url,
            "cache key %s" % nurl,
            "real url %s" % nurl,
            "valid",
        ]
        self.direct(url, resultlines)
        # ftp three slashes
        url = "ftp:///%s:%d/" % (self.host, self.port)
        nurl = self.norm(url)
        resultlines = [
            "url %s" % url,
            "cache key None",
            "real url %s" % nurl,
            "error",
        ]
        self.direct(url, resultlines)
        # directory listing
        url = "ftp://%s:%d/base/" % (self.host, self.port)
        resultlines = [
            "url %s" % url,
            "cache key %s" % url,
            "real url %s" % url,
            "valid",
            "url test.txt",
            "cache key %stest.txt" % url,
            "real url %stest.txt" % url,
            "name test.txt",
            "warning Content size is zero.",
            "valid",
        ]
        self.direct(url, resultlines, recursionlevel=1)
        # file download
        url = "ftp://%s:%d/file.html" % (self.host, self.port)
        resultlines = [
            "url %s" % url,
            "cache key %s" % url,
            "real url %s" % url,
            "valid",
            "url javascript:loadthis()",
            "cache key javascript:loadthis()",
            "real url javascript:loadthis()",
            "name javascript url",
            "info Javascript URL ignored.",
            "valid",
        ]
        self.direct(url, resultlines, recursionlevel=1)



# Copyright (C) 2005-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test cookie routines.
"""

import os
from pathlib import Path

import linkcheck.cookies
import linkcheck.configuration
import linkcheck.director

from . import TestBase


class TestCookies(TestBase):
    """Test cookie routines."""

    def test_cookie_parse_multiple_headers(self):
        lines = [
            "Host: example.org",
            "Path: /hello",
            'Set-cookie: ID="smee"',
            'Set-cookie: spam="egg"',
        ]
        from_headers = linkcheck.cookies.from_headers
        cookies = from_headers("\r\n".join(lines))
        self.assertEqual(len(cookies), 2)
        for cookie in cookies:
            self.assertEqual(cookie.domain, "example.org")
            self.assertEqual(cookie.path, "/hello")
        self.assertEqual(cookies[0].name, "ID")
        self.assertEqual(cookies[0].value, "smee")
        self.assertEqual(cookies[1].name, "spam")
        self.assertEqual(cookies[1].value, "egg")

    def test_cookie_parse_multiple_values(self):
        lines = [
            "Host: example.org",
            'Set-cookie: baggage="elitist"; comment="hologram"',
        ]
        from_headers = linkcheck.cookies.from_headers
        cookies = from_headers("\r\n".join(lines))
        self.assertEqual(len(cookies), 2)
        for cookie in cookies:
            self.assertEqual(cookie.domain, "example.org")
            self.assertEqual(cookie.path, "/")
        self.assertEqual(cookies[0].name, "baggage")
        self.assertEqual(cookies[0].value, "elitist")
        self.assertEqual(cookies[1].name, "comment")
        self.assertEqual(cookies[1].value, "hologram")

    def test_cookie_parse_error(self):
        lines = [
            " Host: imaweevil.org",
            'Set-cookie: baggage="elitist"; comment="hologram"',
        ]
        from_headers = linkcheck.cookies.from_headers
        self.assertRaises(ValueError, from_headers, "\r\n".join(lines))

    def test_cookie_file(self):
        # Regression test for https://github.com/linkchecker/linkchecker/issues/62
        config = linkcheck.configuration.Configuration()
        here = os.path.dirname(__file__)
        config["cookiefile"] = os.path.join(here, "cookies.txt")
        aggregate = linkcheck.director.get_aggregate(config)
        aggregate.add_request_session()
        session = aggregate.get_request_session()
        self.assertEqual({c.name for c in session.cookies}, {"om", "multiple", "are"})

    def test_empty_cookie_file(self):
        self.assertRaises(
            linkcheck.LinkCheckerError,
            linkcheck.cookies.from_file,
            Path(__file__).parent / "configuration/data/config.empty",
        )



# Copyright (C) 2004-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test http checking.
"""
from .httpserver import HttpServerTest, CookieRedirectHttpRequestHandler


class TestHttpsRedirect(HttpServerTest):
    """Test https:// link redirection checking."""

    def __init__(self, methodName="runTest"):
        super().__init__(methodName=methodName)
        self.handler = RedirectHttpsRequestHandler

    def test_redirect(self):
        url = "http://localhost:%d/redirect1" % self.port
        nurl = url
        # rurl = "https://localhost:%d/newurl1" % self.port
        resultlines = [
            "url %s" % url,
            "cache key %s" % nurl,
            "real url %s" % url,
            # XXX the redirect fails because this is not an SSL server
            # "info Redirected to `%s'." % rurl.replace('http:', 'https:'),
            # "valid",
            # "url %s" % rurl,
            # "cache key %s" % rurl,
            # "real url %s" % rurl,
            "error",
        ]
        self.direct(url, resultlines, recursionlevel=0)


class RedirectHttpsRequestHandler(CookieRedirectHttpRequestHandler):
    def redirect(self):
        """Redirect request."""
        path = self.path.replace("redirect", "newurl")
        port = self.server.server_address[1]
        url = "https://localhost:%d%s" % (port, path)
        self.send_response(302)
        self.send_header("Location", url)
        self.end_headers()



#!/usr/bin/env python
"""
View yappi profiling data.

Usage: $0 <filename>
"""
import sys
import yappi


def main(args):
    filename = args[0]
    stats = yappi.YFuncStats()
    stats.add(filename)
    stats.print_all()


if __name__ == "__main__":
    main(sys.argv[1:])



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
A GraphXML logger.
"""

from .xmllog import _XMLLogger
from .graph import _GraphLogger


class GraphXMLLogger(_XMLLogger, _GraphLogger):
    """XML output mirroring the GML structure. Easy to parse with any XML
    tool."""

    LoggerName = 'gxml'

    LoggerArgs = {
        "filename": "linkchecker-out.gxml",
    }

    def __init__(self, **kwargs):
        """Initialize graph node list and internal id counter."""
        args = self.get_args(kwargs)
        super().__init__(**args)
        self.nodes = {}
        self.nodeid = 0

    def start_output(self):
        """Write start of checking info as xml comment."""
        super().start_output()
        self.xml_start_output()
        self.xml_starttag('GraphXML')
        self.xml_starttag('graph', attrs={"isDirected": "true"})
        self.flush()

    def log_url(self, url_data):
        """Write one node and all possible edges."""
        node = self.get_node(url_data)
        if node:
            self.xml_starttag('node', attrs={"name": "%d" % node["id"]})
            self.xml_tag("label", node["label"])
            if self.has_part("realurl"):
                self.xml_tag("url", node["url"])
            self.xml_starttag("data")
            if node["dltime"] >= 0 and self.has_part("dltime"):
                self.xml_tag("dltime", "%f" % node["dltime"])
            if node["size"] >= 0 and self.has_part("dlsize"):
                self.xml_tag("size", "%d" % node["size"])
            if node["checktime"] and self.has_part("checktime"):
                self.xml_tag("checktime", "%f" % node["checktime"])
            if self.has_part("extern"):
                self.xml_tag("extern", "%d" % node["extern"])
            self.xml_endtag("data")
            self.xml_endtag("node")

    def write_edge(self, node):
        """Write one edge."""
        attrs = {
            "source": "%d" % self.nodes[node["parent_url"]]["id"],
            "target": "%d" % node["id"],
        }
        self.xml_starttag("edge", attrs=attrs)
        self.xml_tag("label", node["label"])
        self.xml_starttag("data")
        if self.has_part("result"):
            self.xml_tag("valid", "%d" % node["valid"])
        self.xml_endtag("data")
        self.xml_endtag("edge")

    def end_output(self, **kwargs):
        """Finish graph output, and print end of checking info as xml
        comment."""
        self.xml_endtag("graph")
        self.xml_endtag("GraphXML")
        self.xml_end_output()
        self.close_fileoutput()



# Copyright (C) 2006-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
URL checking functions.
"""
import copy
import time
from . import task
from ..cache import urlqueue
from .. import parser

# Interval in which each check thread looks if it's stopped.
QUEUE_POLL_INTERVALL_SECS = 1.0


def check_urls(urlqueue, logger):
    """Check URLs without threading."""
    while not urlqueue.empty():
        url_data = urlqueue.get()
        try:
            check_url(url_data, logger)
        finally:
            urlqueue.task_done(url_data)


def check_url(url_data, logger):
    """Check a single URL with logging."""
    if url_data.has_result:
        logger.log_url(url_data.to_wire())
    else:
        cache = url_data.aggregate.result_cache
        key = url_data.cache_url
        result = cache.get_result(key)
        if result is None:
            # check
            check_start = time.time()
            try:
                url_data.check()
                do_parse = url_data.check_content()
                url_data.checktime = time.time() - check_start
                # Add result to cache
                result = url_data.to_wire()
                cache.add_result(key, result)
                for alias in url_data.aliases:
                    # redirect aliases
                    cache.add_result(alias, result)
                logger.log_url(result)
                # parse content recursively
                # XXX this could add new warnings which should be cached.
                if do_parse:
                    parser.parse_url(url_data)
            finally:
                # close/release possible open connection
                url_data.close_connection()
        else:
            # copy data from cache and adjust it
            result = copy.copy(result)
            result.parent_url = url_data.parent_url
            result.base_ref = url_data.base_ref or ""
            result.base_url = url_data.base_url or ""
            result.line = url_data.line
            result.column = url_data.column
            result.level = url_data.recursion_level
            result.name = url_data.name
            logger.log_url(result)


class Checker(task.LoggedCheckedTask):
    """URL check thread."""

    def __init__(self, urlqueue, logger, add_request_session):
        """Store URL queue and logger."""
        super().__init__(logger)
        self.urlqueue = urlqueue
        self.origname = self.name
        self.add_request_session = add_request_session

    def run_checked(self):
        """Check URLs in the queue."""
        # construct per-thread HTTP/S requests session
        self.add_request_session()
        while not self.stopped(0):
            self.check_url()

    def check_url(self):
        """Try to get URL data from queue and check it."""
        try:
            url_data = self.urlqueue.get(timeout=QUEUE_POLL_INTERVALL_SECS)
            if url_data is not None:
                try:
                    self.check_url_data(url_data)
                finally:
                    self.urlqueue.task_done(url_data)
                self.name = self.origname
        except urlqueue.Empty:
            pass
        except Exception:
            self.internal_error()

    def check_url_data(self, url_data):
        """Check one URL data instance."""
        self.name = "CheckThread-%s" % (url_data.url or "")
        check_url(url_data, self.logger)



# Copyright (C) 2005-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test linkparser routines.
"""

from linkcheck.htmlutil import htmlsoup, linkparse

from . import TestBase


class TestLinkparser(TestBase):
    """
    Test link parsing.
    """

    def _test_one_link(self, content, url):
        self.count_url = 0
        linkparse.find_links(
            htmlsoup.make_soup(content), self._test_one_url(url), linkparse.LinkTags
        )
        self.assertEqual(self.count_url, 1)

    def _test_one_url(self, origurl):
        """Return parser callback function."""

        def callback(url, line, column, name, base):
            self.count_url += 1
            self.assertEqual(origurl, url)

        return callback

    def _test_no_link(self, content):
        def callback(url, line, column, name, base):
            self.assertTrue(False, "URL %r found" % url)

        linkparse.find_links(htmlsoup.make_soup(content), callback, linkparse.LinkTags)

    def test_href_parsing(self):
        # Test <a href> parsing.
        content = '<a href="%s">'
        url = "alink"
        self._test_one_link(content % url, url)
        url = " alink"
        self._test_one_link(content % url, url)
        url = "alink "
        self._test_one_link(content % url, url)
        url = " alink "
        self._test_one_link(content % url, url)

    def test_rel_parsing(self):
        # Test <link rel> parsing.
        content = '<link rel="%s" href="%s">'
        rel = "dns-prefetch"
        url = "https://alink"
        expected = "dns://alink"
        self._test_one_link(content % (rel, url), expected)
        url = "//alink/"
        self._test_one_link(content % (rel, url), expected)
        rel = "preconnect"
        url = "https://alink"
        self._test_one_link(content % (rel, url), expected)
        rel = "dns-prefetch preconnect"
        url = "https://alink"
        self._test_one_link(content % (rel, url), expected)

    def test_link_without_rel_parsing(self):
        # <link> tags without rel attr should not raise TypeError.
        content = '<link href="%s">'
        url = "https://alink"
        expected = "https://alink"
        # Dummy test, we just have to make sure no error was raised.
        self._test_one_link(content % url, expected)

    def test_img_srcset_parsing(self):
        content = '<img srcset="%s 1x">'
        url = "imagesmall.jpg"
        self._test_one_link(content % url, url)
        content = '<img srcset="imagesmall.jpg 1x,">'
        url = "imagesmall.jpg"
        self._test_one_link(content, url)
        content = '<img srcset="data:image/vnd.microsoft.icon,000001000200">'
        url = "data:image/vnd.microsoft.icon,000001000200"
        self._test_one_link(content, url)

    def test_itemtype_parsing(self):
        content = '<div itemtype="%s">'
        url = "http://example.org/Movie"
        self._test_one_link(content % url, url)

    def test_form_parsing(self):
        # Test <form action> parsing
        content = '<form action="%s">'
        url = "alink"
        self._test_one_link(content % url, url)
        content = '<form action="%s" method="POST">'
        url = "alink"
        self._test_no_link(content % url)

    def test_css_parsing(self):
        # Test css style attribute parsing.
        content = '<table style="background: url(%s) no-repeat" >'
        url = "alink"
        self._test_one_link(content % url, url)
        content = '<table style="background: url(%s) no-repeat" >'
        self._test_one_link(content % url, url)
        content = '<table style="background: url(%s ) no-repeat" >'
        self._test_one_link(content % url, url)
        content = '<table style="background: url( %s ) no-repeat" >'
        self._test_one_link(content % url, url)
        content = "<table style=\"background: url('%s') no-repeat\" >"
        self._test_one_link(content % url, url)
        content = "<table style='background: url(\"%s\") no-repeat' >"
        self._test_one_link(content % url, url)
        content = "<table style=\"background: url('%s' ) no-repeat\" >"
        self._test_one_link(content % url, url)
        content = "<table style='background: url( \"%s\") no-repeat' >"
        self._test_one_link(content % url, url)

    def test_comment_stripping(self):
        strip = linkparse.strip_c_comments
        content = "/* url('http://example.org')*/"
        self.assertEqual(strip(content), "")
        content = "/* * * **/"
        self.assertEqual(strip(content), "")
        content = "/* * /* * **//* */"
        self.assertEqual(strip(content), "")
        content = "a/* */b/* */c"
        self.assertEqual(strip(content), "abc")

    def test_url_quoting(self):
        url = "http://example.com/bla/a=b"
        content = '<a href="%s&quot;">'
        self._test_one_link(content % url, url + '"')



# Copyright (C) 2020 Chris Mayo
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test that <meta name="robots" content="nofollow"> is respected when using http
and ignored when checking a local file.
Also test different values of the content attribute are correctly matched.
"""
import unittest

import linkcheck.configuration
import linkcheck.director
from linkcheck.htmlutil.htmlsoup import make_soup
from . import get_url_from

from . import LinkCheckTest
from .httpserver import HttpServerTest


class TestHttpMetaRobots(HttpServerTest):
    """Test <meta name="robots" content="nofollow"> using http."""

    def test_http_meta_robots(self):
        url = self.get_url("norobots.html")
        resultlines = [
            "url %s" % url,
            "cache key %s" % url,
            "real url %s" % url,
            "valid",
        ]
        self.direct(url, resultlines, recursionlevel=1)


class TestFileMetaRobots(LinkCheckTest):
    """Test <meta name="robots" content="nofollow"> from a file."""

    def test_file_meta_robots(self):
        datapath = "file://%(curdir)s/%(datadir)s/%%s" % self.get_attrs()
        url = datapath % "norobots.html"
        dncurl = datapath % "do_not_check.html"
        resultlines = [
            "url %s" % url,
            "cache key %s" % url,
            "real url %s" % url,
            "valid",
            "url do_not_check.html",
            "cache key %s" % dncurl,
            "real url %s" % dncurl,
            "name bla",
            "error",
        ]
        self.direct(url, resultlines, recursionlevel=1)


class TestMetaRobotsVariants(unittest.TestCase):
    """Test different values of the robots meta directive content attribute"""

    def test_nofollow_variants(self):
        config = linkcheck.configuration.Configuration()
        aggregate = linkcheck.director.get_aggregate(config)
        url = "http://example.org"
        url_data = get_url_from(url, 0, aggregate)
        url_data.content_type = "text/html"

        url_data.soup = make_soup('<meta name="robots" content="nofollow">')
        self.assertFalse(url_data.content_allows_robots())

        url_data.soup = make_soup(
            '<meta name="robots" content="nocache, Nofollow, noimageindex">'
        )
        self.assertFalse(url_data.content_allows_robots())

        url_data.soup = make_soup('<meta name="robots" content="noindex, follow">')
        self.assertTrue(url_data.content_allows_robots())



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Main functions for link parsing
"""
from .. import strformat, url as urlutil
from ..htmlutil import linkparse
from ..bookmarks import firefox


def parse_url(url_data):
    """Parse a URL."""
    if url_data.is_directory():
        # both ftp and file links represent directories as HTML data
        key = "html"
    elif (
        url_data.is_file()
        and firefox.has_sqlite
        and firefox.extension.search(url_data.url)
    ):
        key = "firefox"
    elif url_data.scheme == "itms-services":
        key = "itms_services"
    else:
        # determine parse routine according to content types
        mime = url_data.content_type
        key = url_data.ContentMimetypes[mime]
    funcname = "parse_" + key
    if funcname in globals():
        globals()[funcname](url_data)
    else:
        url_data.aggregate.plugin_manager.run_parser_plugins(url_data, pagetype=key)


def parse_html(url_data):
    """Parse into HTML content and search for URLs to check.
    Found URLs are added to the URL queue.
    """
    linkparse.find_links(url_data.get_soup(), url_data.add_url, linkparse.LinkTags)


def parse_opera(url_data):
    """Parse an opera bookmark file."""
    from ..bookmarks.opera import parse_bookmark_data

    for url, name, lineno in parse_bookmark_data(url_data.get_content()):
        url_data.add_url(url, line=lineno, name=name)


def parse_chromium(url_data):
    """Parse a Chromium or Google Chrome bookmark file."""
    from ..bookmarks.chromium import parse_bookmark_data

    for url, name in parse_bookmark_data(url_data.get_content()):
        url_data.add_url(url, name=name)


def parse_safari(url_data):
    """Parse a Safari bookmark file."""
    from ..bookmarks.safari import parse_bookmark_data

    for url, name in parse_bookmark_data(url_data.get_raw_content()):
        url_data.add_url(url, name=name)


def parse_text(url_data):
    """Parse a text file with one url per line; comment and blank
    lines are ignored."""
    lineno = 0
    for line in url_data.get_content().splitlines():
        lineno += 1
        line = line.strip()
        if not line or line.startswith('#'):
            continue
        url_data.add_url(line, line=lineno)


def parse_css(url_data):
    """
    Parse a CSS file for url() patterns.
    """
    lineno = 0
    linkfinder = linkparse.css_url_re.finditer
    strip_comments = linkparse.strip_c_comments
    for line in strip_comments(url_data.get_content()).splitlines():
        lineno += 1
        for mo in linkfinder(line):
            column = mo.start("url")
            url = strformat.unquote(mo.group("url").strip())
            url_data.add_url(url, line=lineno, column=column)


def parse_swf(url_data):
    """Parse a SWF file for URLs."""
    linkfinder = linkparse.swf_url_re.finditer
    for mo in linkfinder(url_data.get_raw_content()):
        # We're scraping binary data for anything that looks like an URL using
        # a regex that matches only ASCII characters.  Any non-ASCII characters
        # in the URL are expected to be %-encoded.
        url = mo.group().decode('ascii')
        url_data.add_url(url)


def parse_wml(url_data):
    """Parse into WML content and search for URLs to check.
    Found URLs are added to the URL queue.
    """
    linkparse.find_links(url_data.get_soup(), url_data.add_url, linkparse.WmlTags)


def parse_firefox(url_data):
    """Parse a Firefox3 bookmark file."""
    filename = url_data.get_os_filename()
    for url, name in firefox.parse_bookmark_file(filename):
        url_data.add_url(url, name=name)


def parse_itms_services(url_data):
    """Get "url" CGI parameter value as child URL."""
    query = url_data.urlparts[3]
    for k, v, sep in urlutil.parse_qsl(
        query, encoding=url_data.encoding, keep_blank_values=True
    ):
        if k == "url":
            url_data.add_url(v)
            break


from .sitemap import parse_sitemap, parse_sitemapindex  # noqa: F401



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Output logging support for different formats.
"""

import sys
import os
import datetime
import time
import codecs
import abc

from .. import log, LOG_CHECK, strformat, dummy, configuration, i18n

_ = lambda x: x  # noqa: E731
Fields = dict(
    realurl=_("Real URL"),
    cachekey=_("Cache key"),
    result=_("Result"),
    base=_("Base"),
    name=_("Name"),
    parenturl=_("Parent URL"),
    extern=_("Extern"),
    info=_("Info"),
    warning=_("Warning"),
    dltime=_("D/L time"),
    dlsize=_("Size"),
    checktime=_("Check time"),
    url=_("URL"),
    level=_("Level"),
    modified=_("Modified"),
)
del _

ContentTypes = dict(image=0, text=0, video=0, audio=0, application=0, mail=0, other=0)


class LogStatistics:
    """Gather log statistics:
    - number of errors, warnings and valid links
    - type of contents (image, video, audio, text, ...)
    - URL lengths
    """

    def __init__(self):
        """Initialize log statistics."""
        self.reset()

    def reset(self):
        """Reset all log statistics to default values."""
        # number of logged URLs
        self.number = 0
        # number of encountered URL errors
        self.errors = 0
        # number of URL errors that were printed
        self.errors_printed = 0
        # number of URL warnings
        self.warnings = 0
        # number of URL warnings that were printed
        self.warnings_printed = 0
        # number of internal errors
        self.internal_errors = 0
        # link types
        self.link_types = ContentTypes.copy()
        # URL length statistics
        self.max_url_length = 0
        self.min_url_length = 0
        self.avg_url_length = 0.0
        self.avg_number = 0
        # overall downloaded bytes
        self.downloaded_bytes = None

    def log_url(self, url_data, do_print):
        """Log URL statistics."""
        self.number += 1
        if not url_data.valid:
            self.errors += 1
            if do_print:
                self.errors_printed += 1
        num_warnings = len(url_data.warnings)
        self.warnings += num_warnings
        if do_print:
            self.warnings_printed += num_warnings
        if url_data.content_type:
            key = url_data.content_type.split('/', 1)[0].lower()
            if key not in self.link_types:
                key = "other"
        elif url_data.url.startswith("mailto:"):
            key = "mail"
        else:
            key = "other"
        self.link_types[key] += 1
        if url_data.url:
            n = len(url_data.url)
            self.max_url_length = max(n, self.max_url_length)
            if self.min_url_length == 0:
                self.min_url_length = n
            else:
                self.min_url_length = min(n, self.min_url_length)
            # track average number separately since empty URLs do not count
            self.avg_number += 1
            # calculate running average
            self.avg_url_length += (n - self.avg_url_length) / self.avg_number

    def log_internal_error(self):
        """Increase internal error count."""
        self.internal_errors += 1


class _Logger(abc.ABC):
    """
    Base class for logging of checked urls. It defines the public API
    (see below) and offers basic functionality for all loggers.

    Each logger offers the following functions:

    * start_output()
        Initialize and start log output. Most loggers print a comment
        with copyright information.
    * end_output(**kwargs)
        Finish log output, possibly flushing buffers. Most loggers also
        print some statistics.
        Custom keyword arguments can be given for different loggers.
    * log_filter_url(url_data, do_print)
        Log a checked URL. The url_data object is a transport form of
        the UrlData class. The do_print flag indicates if this URL
        should be logged or just used to update internal statistics.

    Each subclassed logger must implement the following functions:

    * start_output()
        Also call the base class implementation of this.
    * end_output(**kwargs)
        See above.
    * log_url(url_data)
        Log a checked URL. Called by log_filter_url if do_print is True.
    """

    # A lowercase name for this logger, usable for option values
    LoggerName = None

    # Default log configuration
    LoggerArgs = {}

    def __init__(self, **args):
        """
        Initialize a logger, looking for part restrictions in kwargs.
        """
        if 'parts' in args and "all" not in args['parts']:
            # only log given parts
            self.logparts = args['parts']
        else:
            # log all parts
            self.logparts = None
        # number of spaces before log parts for alignment
        self.logspaces = {}
        # maximum indent of spaces for alignment
        self.max_indent = 0
        # log statistics
        self.stats = LogStatistics()
        # encoding of output
        encoding = args.get("encoding", i18n.default_encoding)
        try:
            encoding = codecs.lookup(encoding).name
        except LookupError:
            encoding = i18n.default_encoding
        self.output_encoding = encoding
        # how to handle codec errors
        self.codec_errors = "replace"
        # Flag to see if logger is active. Can be deactivated on errors.
        self.is_active = True

    def get_args(self, kwargs):
        """Construct log configuration from default and user args."""
        args = dict(self.LoggerArgs)
        args.update(kwargs)
        return args

    def get_charset_encoding(self):
        """Translate the output encoding to a charset encoding name."""
        if self.output_encoding == "utf-8-sig":
            return "utf-8"
        return self.output_encoding

    def encode(self, s):
        """Encode string with output encoding."""
        assert isinstance(s, str)
        return s.encode(self.output_encoding, self.codec_errors)

    def init_fileoutput(self, args):
        """
        Initialize self.fd file descriptor from args. For file output
        (used when the fileoutput arg is given), the self.fd
        initialization is deferred until the first self.write() call.
        This avoids creation of an empty file when no output is written.
        """
        self.filename = None
        self.close_fd = False
        self.fd = None
        if args.get('fileoutput'):
            self.filename = os.path.expanduser(args['filename'])
        elif 'fd' in args:
            self.fd = args['fd']
        else:
            self.fd = self.create_fd()

    def start_fileoutput(self):
        """Start output to configured file."""
        path = os.path.dirname(self.filename)
        try:
            if path and not os.path.isdir(path):
                os.makedirs(path)
            self.fd = self.create_fd()
            self.close_fd = True
        except OSError:
            msg = sys.exc_info()[1]
            log.warn(
                LOG_CHECK,
                "Could not open file %r for writing: %s\n"
                "Disabling log output of %s",
                self.filename,
                msg,
                self,
            )
            self.fd = dummy.Dummy()
            self.is_active = False
        self.filename = None

    def create_fd(self):
        """Create open file descriptor."""
        if self.filename is None:
            return i18n.get_encoded_writer(
                encoding=self.output_encoding, errors=self.codec_errors
            )
        return codecs.open(self.filename, "wb", self.output_encoding, self.codec_errors)

    def close_fileoutput(self):
        """
        Flush and close the file output denoted by self.fd.
        """
        if self.fd is not None:
            try:
                self.flush()
            except OSError:
                # ignore flush errors
                pass
            if self.close_fd:
                try:
                    self.fd.close()
                except OSError:
                    # ignore close errors
                    pass
            self.fd = None

    def check_date(self):
        """
        Check for special dates.
        """
        now = datetime.date.today()
        if now.day == 7 and now.month == 1:
            msg = _("Happy birthday for LinkChecker, I'm %d years old today!")
            self.comment(msg % (now.year - 2000))

    def comment(self, s, **args):
        """
        Write a comment and a newline. This method just prints
        the given string.
        """
        self.writeln(s=s, **args)

    def wrap(self, lines, width):
        """
        Return wrapped version of given lines.
        """
        sep = os.linesep + os.linesep
        text = sep.join(lines)
        kwargs = dict(
            subsequent_indent=" " * self.max_indent,
            initial_indent=" " * self.max_indent,
            break_long_words=False,
            break_on_hyphens=False,
        )
        return strformat.wrap(text, width, **kwargs).lstrip()

    def write(self, s, **args):
        """Write string to output descriptor. Strips control characters
        from string before writing.
        """
        if self.filename is not None:
            self.start_fileoutput()
        if self.fd is None:
            # Happens when aborting threads times out
            log.warn(LOG_CHECK, "writing to uninitialized or closed file")
        else:
            try:
                self.fd.write(s, **args)
            except OSError:
                msg = sys.exc_info()[1]
                log.warn(
                    LOG_CHECK,
                    "Could not write to output file: %s\n"
                    "Disabling log output of %s",
                    msg,
                    self,
                )
                self.close_fileoutput()
                self.fd = dummy.Dummy()
                self.is_active = False

    def writeln(self, s="", **args):
        """
        Write string to output descriptor plus a newline.
        """
        self.write(f"{s}{os.linesep}", **args)

    def has_part(self, name):
        """
        See if given part name will be logged.
        """
        if self.logparts is None:
            # log all parts
            return True
        return name in self.logparts

    def part(self, name):
        """
        Return translated part name.
        """
        return _(Fields.get(name, ""))

    def spaces(self, name):
        """
        Return indent of spaces for given part name.
        """
        return self.logspaces[name]

    def start_output(self):
        """
        Start log output.
        """
        # map with spaces between part name and value
        if self.logparts is None:
            parts = Fields.keys()
        else:
            parts = self.logparts
        values = (self.part(x) for x in parts)
        # maximum indent for localized log part names
        self.max_indent = max(len(x) for x in values) + 1
        for key in parts:
            numspaces = self.max_indent - len(self.part(key))
            self.logspaces[key] = " " * numspaces
        self.stats.reset()
        self.starttime = time.time()

    def log_filter_url(self, url_data, do_print):
        """
        Log a new url with this logger if do_print is True. Else
        only update accounting data.
        """
        self.stats.log_url(url_data, do_print)
        if do_print:
            self.log_url(url_data)

    def write_intro(self):
        """Write intro comments."""
        self.comment(
            _("created by %(app)s at %(time)s")
            % {"app": configuration.AppName, "time": strformat.strtime(self.starttime)}
        )
        self.comment(
            _("Read the documentation at %(url)s") % {'url': configuration.Url}
        )
        self.comment(
            _("Write comments and bugs to %(url)s") % {'url': configuration.SupportUrl}
        )
        self.check_date()

    def write_outro(self):
        """Write outro comments."""
        self.stoptime = time.time()
        duration = self.stoptime - self.starttime
        self.comment(
            _("Stopped checking at %(time)s (%(duration)s)")
            % {
                "time": strformat.strtime(self.stoptime),
                "duration": strformat.strduration_long(duration),
            }
        )

    @abc.abstractmethod
    def log_url(self, url_data):
        """
        Log a new url with this logger.
        """
        pass

    @abc.abstractmethod
    def end_output(self, **kwargs):
        """
        End of output, used for cleanup (eg output buffer flushing).
        """
        pass

    def __str__(self):
        """
        Return class name.
        """
        return self.__class__.__name__

    def __repr__(self):
        """
        Return class name.
        """
        return repr(self.__class__.__name__)

    def flush(self):
        """
        If the logger has internal buffers, flush them.
        Ignore flush I/O errors since we are not responsible for proper
        flushing of log output streams.
        """
        if hasattr(self, "fd"):
            try:
                self.fd.flush()
            except (OSError, AttributeError):
                pass

    def log_internal_error(self):
        """Indicate that an internal error occurred in the program."""
        log.warn(LOG_CHECK, "internal error occurred")
        self.stats.log_internal_error()

    def format_modified(self, modified, sep=" "):
        """Format modification date in UTC if it's not None.

        @param modified: modification date in UTC
        @type modified: datetime or None
        @return: formatted date or empty string
        @rtype: unicode
        """
        if modified is not None:
            return modified.strftime(f"%Y-%m-%d{sep}%H:%M:%S.%fZ")
        return ""


def _get_loggers():
    """Return list of Logger classes."""
    from .. import loader

    modules = loader.get_package_modules('logger', __path__)
    return list(loader.get_plugins(modules, [_Logger]))


LoggerClasses = _get_loggers()
LoggerNames = [x.LoggerName for x in LoggerClasses]
LoggerKeys = ", ".join(repr(x) for x in LoggerNames)



# Copyright (C) 2003-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Ip number related utility functions.
"""

import ipaddress
import re
import socket
from .. import log, LOG_CHECK


def is_valid_ip(ip):
    """
    Return True if given ip is a valid IPv4 or IPv6 address.
    """
    try:
        ipaddress.ip_address(ip)
    except ValueError:
        return False
    return True


def resolve_host(host):
    """
    Return list of ip numbers for given host.

    @param host: hostname or IP address
    """
    ips = []
    try:
        for res in socket.getaddrinfo(host, None, 0, socket.SOCK_STREAM):
            # res is a tuple (address family, socket type, protocol,
            #  canonical name, socket address)
            # add first ip of socket address
            ips.append(res[4][0])
    except OSError:
        log.info(LOG_CHECK, "Ignored invalid host %r", host)
    return ips


is_obfuscated_ip = re.compile(r"^(0x[a-f0-9]+|[0-9]+)$").match



# Copyright (C) 2022 Chris Mayo
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from pathlib import Path
import re
import shutil
import subprocess

from hatchling.builders.hooks.plugin.interface import BuildHookInterface

try:
    import polib
except ImportError:
    COMPILE_TRANSLATIONS = False
else:
    COMPILE_TRANSLATIONS = True

LOCALE_DIR = ("linkcheck", "data", "locale")
RELEASE_PY = ("linkcheck", "_release.py")


class CustomBuildHook(BuildHookInterface):
    def clean(self, versions):
        Path(*RELEASE_PY).unlink(missing_ok=True)
        shutil.rmtree(str(Path(*LOCALE_DIR)), ignore_errors=True)

    def initialize(self, version, build_data):
        cp = None
        committer_date = committer_year = "unknown"
        try:
            cp = subprocess.run(["git", "log", "-n 1", "HEAD", "--format=%cs"],
                                capture_output=True, check=True, text=True)
        except (FileNotFoundError, subprocess.CalledProcessError):
            # support building wheel from sdist
            if Path(*RELEASE_PY).is_file():
                self.app.display_warning("_release.py already exists")
                return

            try:
                git_archival = Path(".git_archival.txt").read_text()
            except FileNotFoundError:
                self.app.display_warning(".git_archival.txt does not exist")
            else:
                rematch = re.search(r"node-date: ((\d{4})-\d{2}-\d{2})", git_archival)
                if rematch:
                    committer_date = rematch.group(1)
                    committer_year = rematch.group(2)
                else:
                    self.app.display_warning("node-date not substituted")
        else:
            if cp and cp.stdout:
                committer_date = cp.stdout.strip()
                committer_year = committer_date[:4]

        Path(*RELEASE_PY).write_text(f"""\
__app_name__ = "{self.metadata.core.raw_name}"
__version__ = "{self.metadata.version}"
__release_date__ = "{committer_date}"
__copyright_year__ = "{committer_year}"
__author__ = "{self.metadata.core.authors[0]['name']}"
__url__ = "{self.metadata.core.urls["Homepage"]}"
__support_url__ = "{self.metadata.core.urls["Bug Tracker"]}"
""")

        if COMPILE_TRANSLATIONS:
            for po in Path("po").glob("*.po"):
                mo = Path(
                          *LOCALE_DIR,
                          po.stem, "LC_MESSAGES", "linkchecker",
                         ).with_suffix(".mo")
                pofile = polib.pofile(str(po))
                mo.parent.mkdir(exist_ok=True, parents=True)
                pofile.save_as_mofile(str(mo))
        else:
            self.app.display_warning(
                "polib package not found: translations not compiled")



# Copyright (C) 2004-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Define standard test support classes functional for LinkChecker tests.
"""
import os
import re
import difflib
import linkcheck.checker
import linkcheck.configuration
import linkcheck.director
import linkcheck.logger
from .. import get_file, TestBase

# helper alias
get_url_from = linkcheck.checker.get_url_from


class TestLogger(linkcheck.logger._Logger):
    """
    Output logger for automatic regression tests.
    """

    # don't attempt to collect this class because it has an __init__()
    __test__ = False

    LoggerName = "test"

    logparts = [
        "cachekey",
        "realurl",
        "name",
        "base",
        "info",
        "warning",
        "result",
        "url",
    ]

    def __init__(self, **kwargs):
        """
        The kwargs must have "expected" keyword with the expected logger
        output lines.
        """
        args = self.get_args(kwargs)
        args["parts"] = self.logparts
        super().__init__(**args)
        # list of expected output lines
        self.expected = args["expected"]
        # list of real output lines
        self.result = []
        # diff between expected and real output
        self.diff = []

    def normalize(self, result_log):
        # XXX we assume that each log entry has a URL key,
        # maybe we should add an assert into log_url() to that effect?
        # Ensure that log entries are sorted by URL key:
        # - join the result_log items together
        # - split into entries (starting with a URL key)
        # - sort the entries and join together
        # - split the entries back into a list
        return "\n".join(
            sorted(
                [
                    "url %s" % x.strip()
                    for x in re.split(
                        r"^url ",
                        "\n".join(result_log),
                        flags=re.DOTALL | re.MULTILINE,
                    )
                    if x
                ]
            )
        ).splitlines()

    def start_output(self):
        """
        Nothing to do here.
        """
        pass

    def log_url(self, url_data):
        """
        Append logger output to self.result.
        """
        if self.has_part("url"):
            url = "url %s" % url_data.base_url
            self.result.append(url)
        if self.has_part("cachekey"):
            cache_key = url_data.cache_url if url_data.cache_url else None
            self.result.append("cache key %s" % cache_key)
        if self.has_part("realurl"):
            self.result.append("real url %s" % url_data.url)
        if self.has_part("name") and url_data.name:
            self.result.append("name %s" % url_data.name)
        if self.has_part("base") and url_data.base_ref:
            self.result.append("baseurl %s" % url_data.base_ref)
        if self.has_part("info"):
            for info in url_data.info:
                if (
                    "Last modified" not in info
                    and "is located in" not in info
                ):
                    self.result.append("info %s" % info)
        if self.has_part("warning"):
            for tag, warning in url_data.warnings:
                self.result.append("warning %s" % warning)
        if self.has_part("result"):
            self.result.append("valid" if url_data.valid else "error")
        if self.has_part("line"):
            self.result.append("line %s" % url_data.line)
        if self.has_part("col"):
            self.result.append("col %s" % url_data.column)
        if self.has_part("size"):
            self.result.append("size %s" % url_data.size)
        if self.has_part("parent_url"):
            self.result.append("parent_url %s" % url_data.parent_url)
        if self.has_part("page"):
            self.result.append("page %s" % url_data.page)
        if self.has_part("modified"):
            self.result.append("modified %s" % url_data.modified)
        if self.has_part("content_type"):
            self.result.append("content_type %s" % url_data.content_type)
        # note: do not append url_data.result since this is
        # platform dependent

    def end_output(self, linknumber=-1, **kwargs):
        """
        Stores differences between expected and result in self.diff.
        """
        self.expected = self.normalize(self.expected)
        self.result = self.normalize(self.result)
        self.diff = list(
            difflib.unified_diff(
                self.expected,
                self.result,
                fromfile="expected",
                tofile="result",
                lineterm="",
            )
        )


def get_file_url(filename):
    return re.sub("^([a-zA-Z]):", r"/\1|", filename.replace("\\", "/"))


def add_fileoutput_config(config):
    if os.name == "posix":
        devnull = "/dev/null"
    elif os.name == "nt":
        devnull = "NUL"
    else:
        return
    for ftype in linkcheck.logger.LoggerNames:
        if ftype in ("test", "failures"):
            continue
        logger = config.logger_new(ftype, fileoutput=1, filename=devnull)
        config["fileoutput"].append(logger)


def get_test_aggregate(confargs, logargs, logger=TestLogger):
    """Initialize a test configuration object."""
    config = linkcheck.configuration.Configuration()
    config.logger_add(logger)
    config["recursionlevel"] = 1
    config["logger"] = config.logger_new(logger.LoggerName, **logargs)
    add_fileoutput_config(config)
    config["verbose"] = True
    config["threads"] = 0
    config["status"] = False
    config["checkextern"] = True
    config.update(confargs)
    config.sanitize()
    return linkcheck.director.get_aggregate(config)


class LinkCheckTest(TestBase):
    """
    Functional test class with ability to test local files.
    """

    logger = TestLogger

    def norm(self, url, encoding="utf-8"):
        """Helper function to norm a url."""
        return linkcheck.url.url_norm(url, encoding=encoding)[0]

    def get_attrs(self, **kwargs):
        """Return current and data directory as dictionary.
        You can augment the dict with keyword attributes."""
        d = {
            "curdir": get_file_url(os.getcwd()),
            "datadir": "tests/checker/data",
        }
        d.update(kwargs)
        return d

    def get_resultlines(self, filename):
        """
        Return contents of file, as list of lines without line endings,
        ignoring empty lines and lines starting with a hash sign (#).
        """
        resultfile = get_file("%s.result" % filename)
        d = {
            "curdir": get_file_url(os.getcwd()),
            "datadir": get_file_url(get_file()),
        }
        # the webserver uses the first free port number
        if hasattr(self, "port"):
            d["port"] = self.port
        # all result files are encoded in utf-8
        with open(resultfile, encoding="utf-8") as f:
            return [
                line.rstrip("\r\n") % d
                for line in f
                if line.strip() and not line.startswith("#")
            ]

    def get_url(self, filename):
        """Get URL for given filename."""
        return get_file(filename)

    def file_test(self, filename, confargs=None):
        """Check <filename> with expected result in <filename>.result."""
        url = self.get_url(filename)
        if confargs is None:
            confargs = {}
        logargs = {"expected": self.get_resultlines(filename)}
        aggregate = get_test_aggregate(confargs, logargs, logger=self.logger)
        url_data = get_url_from(url, 0, aggregate, extern=(0, 0))
        aggregate.urlqueue.put(url_data)
        linkcheck.director.check_urls(aggregate)
        logger = aggregate.config["logger"]
        diff = logger.diff
        if diff:
            msg = os.linesep.join([url] + diff)
            self.fail(msg)
        if logger.stats.internal_errors:
            self.fail("%d internal errors occurred!" % logger.stats.internal_errors)

    def direct(
        self,
        url,
        resultlines,
        parts=None,
        recursionlevel=0,
        confargs=None,
        url_encoding=None,
    ):
        """Check url with expected result."""
        assert isinstance(url, str), repr(url)
        if confargs is None:
            confargs = {"recursionlevel": recursionlevel}
        else:
            confargs["recursionlevel"] = recursionlevel
        logargs = {"expected": resultlines}
        if parts is not None:
            logargs["parts"] = parts
        aggregate = get_test_aggregate(confargs, logargs)
        # initial URL has recursion level zero
        url_reclevel = 0
        url_data = get_url_from(url, url_reclevel, aggregate, url_encoding=url_encoding)
        aggregate.urlqueue.put(url_data)
        linkcheck.director.check_urls(aggregate)
        logger = aggregate.config["logger"]
        diff = logger.diff
        if diff:
            d = ["Differences found testing %s" % url]
            d.extend(x.rstrip() for x in diff[2:])
            self.fail(os.linesep.join(d))
        if logger.stats.internal_errors:
            self.fail("%d internal errors occurred!" % logger.stats.internal_errors)


class MailTest(LinkCheckTest):
    """Test mailto: link checking."""

    def mail_valid(self, addr, **kwargs):
        """Test valid mail address."""
        return self.mail_test(addr, "valid", **kwargs)

    def mail_error(self, addr, **kwargs):
        """Test error mail address."""
        return self.mail_test(addr, "error", **kwargs)

    def mail_test(self, addr, result, encoding="utf-8", cache_key=None, warning=None):
        """Test mail address."""
        url = self.norm(addr, encoding=encoding)
        if cache_key is None:
            cache_key = url
        resultlines = [
            "url %s" % url,
            "cache key %s" % cache_key,
            "real url %s" % url,
        ]
        if warning:
            resultlines.append("warning %s" % warning)
        resultlines.append(result)
        self.direct(url, resultlines)



# -*- coding: iso-8859-1 -*-
# Copyright (C) 2004-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test url routines.
"""
from . import need_posix, need_windows
import unittest
import os

import linkcheck.url

# 'ftp://user:pass@ftp.foo.net/foo/bar':
#     'ftp://user:pass@ftp.foo.net/foo/bar',
# 'http://USER:pass@www.Example.COM/foo/bar':
#     'http://USER:pass@www.example.com/foo/bar',
# '-':                             '-',

# All portions of the URI must be utf-8 encoded NFC form Unicode strings
# valid:  http://example.com/?q=%C3%87   (C-cedilla U+00C7)
# valid: http://example.com/?q=%E2%85%A0 (Roman numeral one U+2160)
# invalid: http://example.com/?q=%C7      (C-cedilla ISO-8859-1)
# invalid: http://example.com/?q=C%CC%A7
#         (Latin capital letter C + Combining cedilla U+0327)


def url_norm(url, encoding="utf-8"):
    return linkcheck.url.url_norm(url, encoding=encoding)[0]


class TestUrl(unittest.TestCase):
    """Test url norming and quoting."""

    def urlnormtest(self, url, nurl, encoding=None):
        self.assertFalse(
            linkcheck.url.url_needs_quoting(nurl),
            "Result URL %r must not need quoting" % nurl,
        )
        nurl1 = url_norm(url, encoding=encoding)
        self.assertFalse(
            linkcheck.url.url_needs_quoting(nurl1), "Normed URL %r needs quoting" % nurl
        )
        self.assertEqual(nurl1, nurl)

    def test_wayback(self):
        self.assertFalse("http%3A/x" in url_norm("https://a.b.c/*/http://x.y.z"))
        self.assertTrue("http://x" in url_norm("https://a.b.c/*/http://x.y.z"))

    def test_pathattack(self):
        # Windows winamp path attack prevention.
        url = "http://server/..%5c..%5c..%5c..%5c..%5c..%5c..%5c..%5ccskin.zip"
        nurl = "http://server/cskin.zip"
        self.assertEqual(
            linkcheck.url.url_quote(url_norm(url), encoding="iso-8859-1"), nurl
        )

    def test_url_quote(self):
        def url_quote(url):
            return linkcheck.url.url_quote(url, encoding="utf-8")

        url = "http://a:80/bcd"
        self.assertEqual(url_quote(url), url)
        url = "http://a:80/bcd?"
        url2 = "http://a:80/bcd"
        self.assertEqual(url_quote(url), url2)
        url = "http://a:80/bcd?a=b"
        url2 = "http://a:80/bcd?a=b"
        self.assertEqual(url_quote(url), url2)
        url = "a/b"
        self.assertEqual(url_quote(url), url)
        url = "bcd?"
        url2 = "bcd"
        self.assertEqual(url_quote(url), url2)
        url = "bcd?a=b"
        url2 = "bcd?a=b"
        self.assertEqual(url_quote(url), url2)

    def test_norm_quote(self):
        # Test url norm quoting.
        url = (
            "http://groups.google.com/groups?hl=en&lr&ie=UTF-8&"
            "threadm=3845B54D.E546F9BD%40monmouth.com&rnum=2&"
            "prev=/groups%3Fq%3Dlogitech%2Bwingman%2Bextreme%2Bdigital"
            "%2B3d%26hl%3Den%26lr%3D%26ie%3DUTF-8%26selm%3D3845B54D.E5"
            "46F9BD%2540monmouth.com%26rnum%3D2"
        )
        self.urlnormtest(url, url)
        url = "http://redirect.alexa.com/redirect?http://www.offeroptimizer.com"
        self.urlnormtest(url, url)
        url = "http://www.lesgensducinema.com/photo/Philippe%20Nahon.jpg"
        self.urlnormtest(url, url)
        # Only perform percent-encoding where it is essential.
        url = "http://example.com/%7Ejane"
        nurl = "http://example.com/~jane"
        self.urlnormtest(url, nurl)
        url = "http://example.com/%7ejane"
        self.urlnormtest(url, nurl)
        # Always use uppercase A-through-F characters when percent-encoding.
        url = "http://example.com/?q=1%2a2"
        nurl = "http://example.com/?q=1%2A2"
        self.urlnormtest(url, nurl)
        # the no-quote chars
        url = "http://example.com/a*+-();b"
        self.urlnormtest(url, url)
        url = (
            "http://linkchecker.git.sourceforge.net/git/gitweb.cgi"
            "?p=linkchecker/linkchecker;a=blob;f=doc/changelog.txt;hb=HEAD"
        )
        self.urlnormtest(url, url)
        url = "http://www.company.com/path/doc.html?url=/path2/doc2.html?foo=bar"
        self.urlnormtest(url, url)
        url = "http://example.com/#a b"
        nurl = "http://example.com/#a%20b"
        self.urlnormtest(url, nurl)
        url = "http://example.com/?u=http://example2.com?b=c "
        nurl = "http://example.com/?u=http://example2.com?b=c%20"
        self.urlnormtest(url, nurl)
        url = "http://example.com/?u=http://example2.com?b="
        nurl = "http://example.com/?u=http://example2.com?b="
        self.urlnormtest(url, nurl)
        url = "http://localhost:8001/?quoted=ü"
        nurl = "http://localhost:8001/?quoted=%C3%BC"
        self.urlnormtest(url, nurl)
        url = "http://host/?a=b/c+d="
        nurl = "http://host/?a=b/c%20d%3D"
        self.urlnormtest(url, nurl)

    def test_norm_case_sensitivity(self):
        # Test url norm case sensitivity.
        # Always provide the URI scheme in lowercase characters.
        url = "HTTP://example.com/"
        nurl = "http://example.com/"
        self.urlnormtest(url, nurl)
        # Always provide the host, if any, in lowercase characters.
        url = "http://EXAMPLE.COM/"
        nurl = "http://example.com/"
        self.urlnormtest(url, nurl)
        url = "http://EXAMPLE.COM:55/"
        nurl = "http://example.com:55/"
        self.urlnormtest(url, nurl)

    def test_norm_defaultport(self):
        # Test url norm default port recognition.
        # For schemes that define a port, use an empty port if the default
        # is desired
        url = "http://example.com:80/"
        nurl = "http://example.com/"
        self.urlnormtest(url, nurl)
        url = "http://example.com:8080/"
        nurl = "http://example.com:8080/"
        self.urlnormtest(url, nurl)

    def test_norm_host_dot(self):
        # Test url norm host dot removal.
        url = "http://example.com./"
        nurl = "http://example.com/"
        self.urlnormtest(url, nurl)
        url = "http://example.com.:81/"
        nurl = "http://example.com:81/"
        self.urlnormtest(url, nurl)

    def test_norm_fragment_empty(self):
        # Test url norm fragment preserving.
        # Empty fragment identifiers must be preserved:
        url = "http://www.w3.org/2000/01/rdf-schema#"
        nurl = url
        self.urlnormtest(url, nurl)

    def test_norm_fragment(self):
        url = "http://example.org/foo/ #a=1,2,3"
        nurl = "http://example.org/foo/%20#a=1,2,3"
        self.urlnormtest(url, nurl)

    def test_norm_empty_path(self):
        # Test url norm empty path handling.
        # For schemes that define an empty path to be equivalent to a
        # path of "/", use "/".
        url = "http://example.com"
        nurl = "http://example.com"
        self.urlnormtest(url, nurl)
        url = "http://example.com?a=b"
        nurl = "http://example.com/?a=b"
        self.urlnormtest(url, nurl)
        url = "http://example.com#foo"
        nurl = "http://example.com/#foo"
        self.urlnormtest(url, nurl)

    def test_norm_path_backslashes(self):
        # Test url norm backslash path handling.
        # note: this is not RFC conform (see url.py for more info)
        url = r"http://example.com\test.html"
        nurl = "http://example.com/test.html"
        self.urlnormtest(url, nurl)
        url = r"http://example.com/a\test.html"
        nurl = "http://example.com/a/test.html"
        self.urlnormtest(url, nurl)
        url = r"http://example.com\a\test.html"
        nurl = "http://example.com/a/test.html"
        self.urlnormtest(url, nurl)
        url = r"http://example.com\a/test.html"
        nurl = "http://example.com/a/test.html"
        self.urlnormtest(url, nurl)

    def test_norm_path_slashes(self):
        # Test url norm slashes in path handling.
        # reduce duplicate slashes
        url = "http://example.com//a/test.html"
        nurl = "http://example.com/a/test.html"
        self.urlnormtest(url, nurl)
        url = "http://example.com//a/b/"
        nurl = "http://example.com/a/b/"
        self.urlnormtest(url, nurl)

    def test_norm_path_dots(self):
        # Test url norm dots in path handling.
        # Prevent dot-segments appearing in non-relative URI paths.
        url = "http://example.com/a/./b"
        nurl = "http://example.com/a/b"
        self.urlnormtest(url, nurl)
        url = "http://example.com/a/../a/b"
        self.urlnormtest(url, nurl)
        url = "http://example.com/../a/b"
        self.urlnormtest(url, nurl)

    def test_norm_path_relative_dots(self):
        # Test url norm relative path handling with dots.
        # normalize redundant path segments
        url = "/foo/bar/."
        nurl = "/foo/bar/"
        self.urlnormtest(url, nurl)
        url = "/foo/bar/./"
        nurl = "/foo/bar/"
        self.urlnormtest(url, nurl)
        url = "/foo/bar/.."
        nurl = "/foo/"
        self.urlnormtest(url, nurl)
        url = "/foo/bar/../"
        nurl = "/foo/"
        self.urlnormtest(url, nurl)
        url = "/foo/bar/../baz"
        nurl = "/foo/baz"
        self.urlnormtest(url, nurl)
        url = "/foo/bar/../.."
        nurl = "/"
        self.urlnormtest(url, nurl)
        url = "/foo/bar/../../"
        nurl = "/"
        self.urlnormtest(url, nurl)
        url = "/foo/bar/../../baz"
        nurl = "/baz"
        self.urlnormtest(url, nurl)
        url = "/foo/bar/../../../baz"
        nurl = "/baz"
        self.urlnormtest(url, nurl)
        url = "/foo/bar/../../../../baz"
        nurl = "/baz"
        self.urlnormtest(url, nurl)
        url = "/./foo"
        nurl = "/foo"
        self.urlnormtest(url, nurl)
        url = "/../foo"
        nurl = "/foo"
        self.urlnormtest(url, nurl)
        url = "/foo."
        nurl = url
        self.urlnormtest(url, nurl)
        url = "/.foo"
        nurl = url
        self.urlnormtest(url, nurl)
        url = "/foo.."
        nurl = url
        self.urlnormtest(url, nurl)
        url = "/..foo"
        nurl = url
        self.urlnormtest(url, nurl)
        url = "/./../foo"
        nurl = "/foo"
        self.urlnormtest(url, nurl)
        url = "/./foo/."
        nurl = "/foo/"
        self.urlnormtest(url, nurl)
        url = "/foo/./bar"
        nurl = "/foo/bar"
        self.urlnormtest(url, nurl)
        url = "/foo/../bar"
        nurl = "/bar"
        self.urlnormtest(url, nurl)
        url = "../../../images/miniXmlButton.gif"
        nurl = url
        self.urlnormtest(url, nurl)
        url = "/a..b/../images/miniXmlButton.gif"
        nurl = "/images/miniXmlButton.gif"
        self.urlnormtest(url, nurl)
        url = "/.a.b/../foo/"
        nurl = "/foo/"
        self.urlnormtest(url, nurl)
        url = "/..a.b/../foo/"
        nurl = "/foo/"
        self.urlnormtest(url, nurl)
        url = "b/../../foo/"
        nurl = "../foo/"
        self.urlnormtest(url, nurl)
        url = "./foo"
        nurl = "foo"
        self.urlnormtest(url, nurl)

    def test_norm_path_relative_slashes(self):
        # Test url norm relative path handling with slashes.
        url = "/foo//"
        nurl = "/foo/"
        self.urlnormtest(url, nurl)
        url = "/foo///bar//"
        nurl = "/foo/bar/"
        self.urlnormtest(url, nurl)

    def test_mail_url(self):
        # Test mailto URLs.
        # no netloc and no path
        url = "mailto:"
        nurl = url
        self.urlnormtest(url, nurl)
        # standard email
        url = "mailto:user@www.example.org"
        nurl = url
        self.urlnormtest(url, nurl)
        # emails with subject
        url = "mailto:user@www.example.org?subject=a_b"
        nurl = url
        self.urlnormtest(url, nurl)
        url = (
            "mailto:business.inquiries@designingpatterns.com?subject=Business%20Inquiry"
        )
        nurl = url
        self.urlnormtest(url, nurl)

    def test_norm_other(self):
        # Test norming of other schemes.
        url = "news:"
        nurl = "news:"
        self.urlnormtest(url, nurl)
        url = "snews:"
        nurl = "snews://"
        self.urlnormtest(url, nurl)
        url = "nntp:"
        nurl = "nntp://"
        self.urlnormtest(url, nurl)
        url = "news:!$%&/()="
        nurl = "news:!%24%25%26/()="
        self.urlnormtest(url, nurl)
        url = "news:comp.infosystems.www.servers.unix"
        nurl = url
        self.urlnormtest(url, nurl)
        # javascript url
        url = "javascript:loadthis()"
        nurl = url
        self.urlnormtest(url, nurl)
        # ldap url
        url = "ldap://[2001:db8::7]/c=GB?objectClass?one"
        nurl = "ldap://%5B2001:db8::7%5D/c=GB?objectClass?one"
        self.urlnormtest(url, nurl)
        url = "tel:+1-816-555-1212"
        nurl = url
        self.urlnormtest(url, nurl)
        url = "urn:oasis:names:specification:docbook:dtd:xml:4.1.2"
        nurl = "urn:oasis%3Anames%3Aspecification%3Adocbook%3Adtd%3Axml%3A4.1.2"
        self.urlnormtest(url, nurl)

    def test_norm_with_auth(self):
        # Test norming of URLs with authentication tokens.
        url = "telnet://User@www.example.org"
        nurl = url
        self.urlnormtest(url, nurl)
        url = "telnet://User:Pass@www.example.org"
        nurl = url
        self.urlnormtest(url, nurl)
        url = "http://User:Pass@www.example.org/"
        nurl = url
        self.urlnormtest(url, nurl)

    @need_posix
    def test_norm_file1(self):
        url = "file:///a/b.txt"
        nurl = url
        self.urlnormtest(url, nurl)

    @need_windows
    def test_norm_file2(self):
        url = "file:///C|/a/b.txt"
        nurl = url
        self.urlnormtest(url, nurl)

    @need_posix
    def test_norm_file_unicode(self):
        url = "file:///a/b.txt"
        nurl = url
        self.urlnormtest(url, nurl)
        url = "file:///a/ä.txt"
        nurl = "file:///a/%C3%A4.txt"
        self.urlnormtest(url, nurl)
        # url = "file:///\u041c\u043e\u0448\u043a\u043e\u0432\u0430.bin"
        # nurl = "file:///a.bin" # XXX
        # self.urlnormtest(url, nurl)

    def test_norm_invalid(self):
        url = "äöü?:"
        nurl = "%C3%A4%C3%B6%C3%BC?:"
        self.urlnormtest(url, nurl)

    def test_valid(self):
        # Test url validity functions.
        u = "http://www.example.com"
        self.assertTrue(linkcheck.url.is_safe_url(u), u)
        u = "http://www.example.com/"
        self.assertTrue(linkcheck.url.is_safe_url(u), u)
        u = "http://www.example.com/~calvin"
        self.assertTrue(linkcheck.url.is_safe_url(u), u)
        u = "http://www.example.com/a,b"
        self.assertTrue(linkcheck.url.is_safe_url(u), u)
        u = "http://www.example.com#anchor55"
        self.assertTrue(linkcheck.url.is_safe_url(u), u)

    def test_needs_quoting(self):
        # Test url quoting necessity.
        url = "mailto:<calvin@example.org>?subject=Halli Hallo"
        self.assertTrue(linkcheck.url.url_needs_quoting(url), repr(url))
        url = " http://www.example.com/"
        self.assertTrue(linkcheck.url.url_needs_quoting(url), repr(url))
        url = "http://www.example.com/ "
        self.assertTrue(linkcheck.url.url_needs_quoting(url), repr(url))
        url = "http://www.example.com/\n"
        self.assertTrue(linkcheck.url.url_needs_quoting(url), repr(url))
        url = "\nhttp://www.example.com/"
        self.assertTrue(linkcheck.url.url_needs_quoting(url), repr(url))
        url = "http://www.example.com/#a!"
        self.assertTrue(not linkcheck.url.url_needs_quoting(url), repr(url))
        url = "http://www.example.com/#a b"
        self.assertTrue(linkcheck.url.url_needs_quoting(url), repr(url))

    def test_absolute_url(self):
        url = "hutzli:"
        self.assertTrue(linkcheck.url.url_is_absolute(url), repr(url))
        url = "file:/"
        self.assertTrue(linkcheck.url.url_is_absolute(url), repr(url))
        url = ":"
        self.assertTrue(not linkcheck.url.url_is_absolute(url), repr(url))
        url = "/a/b?http://"
        self.assertTrue(not linkcheck.url.url_is_absolute(url), repr(url))

    def test_nopathquote_chars(self):
        if os.name == "nt":
            url = "file:///c|/msys/"
            nurl = url
            self.assertEqual(url_norm(url), nurl)
            self.assertTrue(not linkcheck.url.url_needs_quoting(url))
        url = "http://hulla/a/b/!?c=d"
        nurl = url
        self.assertEqual(url_norm(url), nurl)

    def test_idn_encoding(self):
        # Test idna encoding.
        url = "www.öko.de"
        idna_encode = linkcheck.url.idna_encode
        encurl, is_idn = idna_encode(url)
        self.assertTrue(is_idn)
        self.assertTrue(encurl)
        url = ""
        encurl, is_idn = idna_encode(url)
        self.assertFalse(is_idn)
        self.assertFalse(encurl)
        url = "ä.."
        self.assertRaises(UnicodeError, idna_encode, url)

    def test_splitparam(self):
        # Path parameter split test.
        p = [
            ("", ("", "")),
            ("/", ("/", "")),
            ("a", ("a", "")),
            ("a;", ("a", "")),
            ("a/b;c/d;e", ("a/b;c/d", "e")),
        ]
        for x in p:
            self._splitparam(x)

    def _splitparam(self, x):
        self.assertEqual(linkcheck.url.splitparams(x[0]), (x[1][0], x[1][1]))

    def test_cgi_split(self):
        # Test cgi parameter splitting.
        u = "scid=kb;en-us;Q248840"
        self.assertEqual(linkcheck.url.url_parse_query(u, encoding="utf-8"), u)
        u = "scid=kb;en-us;Q248840&b=c;hulla=bulla"
        self.assertEqual(linkcheck.url.url_parse_query(u, encoding="utf-8"), u)

    def test_long_cgi(self):
        u = "/test%s;" % ("?a=" * 1000)
        self.assertEqual(linkcheck.url.url_parse_query(u, encoding="utf-8"), u)

    def test_port(self):
        is_numeric_port = linkcheck.url.is_numeric_port
        self.assertTrue(is_numeric_port("80"))
        self.assertTrue(is_numeric_port("1"))
        self.assertFalse(is_numeric_port("0"))
        self.assertFalse(is_numeric_port("66000"))
        self.assertFalse(is_numeric_port("-1"))
        self.assertFalse(is_numeric_port("a"))

    def test_safe_domain(self):
        is_safe_domain = linkcheck.url.is_safe_domain
        self.assertFalse(is_safe_domain("a..example.com"))
        self.assertFalse(is_safe_domain("a_b.example.com"))
        self.assertTrue(is_safe_domain("a-b.example.com"))
        self.assertTrue(is_safe_domain("x1.example.com"))

    def test_splitport(self):
        splitport = linkcheck.url.splitport
        netloc = "hostname"
        host, port = splitport(netloc, 99)
        self.assertEqual(host, netloc)
        self.assertEqual(port, 99)
        netloc = "hostname:"
        host, port = splitport(netloc, 99)
        self.assertEqual(host, "hostname")
        self.assertEqual(port, 99)
        netloc = "hostname:42"
        host, port = splitport(netloc, 99)
        self.assertEqual(host, "hostname")
        self.assertEqual(port, 42)
        netloc = "hostname:foo"
        host, port = splitport(netloc, 99)
        self.assertEqual(host, netloc)
        self.assertEqual(port, 99)



# Copyright (C) 2004-2012 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test html parsing.
"""

from linkcheck.htmlutil import htmlsoup

from io import StringIO

import pytest

from .htmllib import pretty_print_html

# list of tuples
# (<test pattern>, <expected parse output>)
parsetests = [
    # start tags
    ("""<a  b="c" >""", """<a b="c"/>"""),
    ("""<a  b='c' >""", """<a b="c"/>"""),
    ("""<a  b=c" >""", """<a b="c&quot;"/>"""),
    ("""<a  b=c' >""", """<a b="c'"/>"""),
    ("""<a  b="" >""", """<a b=""/>"""),
    ("""<a  b='' >""", """<a b=""/>"""),
    ("""<a  b=>""", """<a b=""/>"""),
    ("""<a  b= >""", """<a b=""/>"""),
    ("""<a  =c>""", """<a =c=""/>"""),
    ("""<a  =c >""", """<a =c=""/>"""),
    ("""<a  =>""", """<a ==""/>"""),
    ("""<a  = >""", """<a ==""/>"""),
    ("""<a  b= "c" >""", """<a b="c"/>"""),
    ("""<a  b ="c" >""", """<a b="c"/>"""),
    ("""<a  b = "c" >""", """<a b="c"/>"""),
    ("""<a >""", """<a/>"""),
    ("""<>""", """"""),
    ("""< >""", """"""),
    ("""<aä>""", """<aä/>"""),
    ("""<a aä="b">""", """<a aä="b"/>"""),
    ("""<a a="bä">""", """<a a="b&#228;"/>"""),
    # multiple attribute names should be ignored...
    ("""<a b="c" b="c" >""", """<a b="c"/>"""),
    # ... but which one wins - in our implementation the last one
    ("""<a b="c" b="d" >""", """<a b="d"/>"""),
    # reduce test
    ("""<a  b="c"><""", """<a b="c"><</a>"""),
    # numbers in tag
    ("""<h1>bla</h1>""", """<h1>bla</h1>"""),
    # more start tags
    ("""<a  b=c"><a b="c">""", """<a b="c&quot;"/><a b="c"/>"""),
    ("""<a  b=/c/></a><br>""", """<a b="/c/"/><br/>"""),
    ("""<br/>""", """<br/>"""),
    ("""<a  b="50%"><br>""", """<a b="50%"/><br/>"""),
    # start and end tag (HTML doctype assumed)
    ("""<a/>""", """<a/>"""),
    ("""<meta/>""", """<meta/>"""),
    ("""<MetA/>""", """<meta/>"""),
    # line continuation (Dr. Fun webpage)
    ("""<img bo\\\nrder=0 >""", """<img bo\\="" rder="0"/>"""),
    ("""<img align="mid\\\ndle">""", """<img align="mid\\\ndle"/>"""),
    ("""<img align='mid\\\ndle'>""", """<img align="mid\\\ndle"/>"""),
    # href with $
    ("""<a href="123$456">""", """<a href="123$456"/>"""),
    # quoting
    ("""<a  href=/ >""", """<a href="/"/>"""),
    ("""<a  href= />""", """<a href="/"/>"""),
    ("""<a  href= >""", """<a href=""/>"""),
    ("""<a  href="'" >""", """<a href="'"/>"""),
    ("""<a  href='"' >""", """<a href="&quot;"/>"""),
    ("""<a  href="bla" %]" >""", """<a %]"="" href="bla"/>"""),
    ("""<a  href=bla" >""", """<a href="bla&quot;"/>"""),
    (
        """<a onmouseover=blubb('nav1','',"""
        """'/images/nav.gif',1);move(this); b="c">""",
        """<a b="c" onmouseover="blubb('nav1','',"""
        """'/images/nav.gif',1);move(this);"/>""",
    ),
    (
        """<a onClick=location.href('/index.htm') b="c">""",
        """<a b="c" onclick="location.href('/index.htm')"/>""",
    ),
    # entity resolving
    ("""<a  href="&#6D;ailto:" >""", """<a href="D;ailto:"/>"""),
    ("""<a  href="&amp;ailto:" >""", """<a href="&amp;ailto:"/>"""),
    ("""<a  href="&amp;amp;ailto:" >""", """<a href="&amp;amp;ailto:"/>"""),
    ("""<a  href="&hulla;ailto:" >""", """<a href="&amp;hulla;ailto:"/>"""),
    ("""<a  href="&#109;ailto:" >""", """<a href="mailto:"/>"""),
    ("""<a  href="&#x6D;ailto:" >""", """<a href="mailto:"/>"""),
    # note that \u8156 is not valid encoding and therefore gets removed
    ("""<a  href="&#8156;ailto:" >""", """<a href="&#8156;ailto:"/>"""),
    # mailto link
    (
        """<a  href=mailto:calvin@LocalHost?subject=Hallo&to=michi>1</a>""",
        """<a href="mailto:calvin@LocalHost?subject=Hallo&amp;to=michi">1</a>""",
    ),
    # meta tag with charset encoding
    (
        """<meta http-equiv="content-type" content>""",
        """<meta content="" http-equiv="content-type"/>""",
    ),
    (
        """<meta http-equiv="content-type" content=>""",
        """<meta content="" http-equiv="content-type"/>""",
    ),
    (
        """<meta http-equiv="content-type" content="hulla">""",
        """<meta content="hulla" http-equiv="content-type"/>""",
    ),
    (
        """<meta http-equiv="content-type" content="text/html; charset=iso8859-1">""",
        """<meta content="text/html; charset=iso8859-1" http-equiv="content-type"/>""",
    ),
    (
        """<meta http-equiv="content-type" content="text/html; charset=hulla">""",
        """<meta content="text/html; charset=hulla" http-equiv="content-type"/>""",
    ),
    # missing > in end tag
    ("""</td <td  a="b" >""", """"""),
    ("""</td<td  a="b" >""", """"""),
    # missing beginning quote
    ("""<td a=b">""", """<td a="b&quot;"/>"""),
    # stray < before start tag
    ("""<0.<td  a="b" >""", """<td a="b"/>"""),
    # HTML5 tags
    ("""<audio  src=bla>""", """<audio src="bla"/>"""),
    ("""<button  formaction=bla>""", """<button formaction="bla"/>"""),
    ("""<html  manifest=bla>""", """<html manifest="bla"/>"""),
    ("""<source  src=bla>""", """<source src="bla"/>"""),
    ("""<track  src=bla>""", """<track src="bla"/>"""),
    ("""<video  src=bla>""", """<video src="bla"/>"""),
    # Test inserted tag s
    ("""<a></a><b></b>""", """<a/><b/>"""),
    # This is not correct result for an HTML parser, but it is for us
    ("""<b><a></a></b>""", """<b/><a/>"""),
]


class TestParser:
    """
    Test html parser.
    """
    @pytest.mark.parametrize("_in, _out", parsetests)
    def test_parse(self, _in, _out):
        # Parse all test patterns in one go.
        out = StringIO()
        pretty_print_html(out, htmlsoup.make_soup(_in))
        self.check_results(_in, _out, out)

    def check_results(self, _in, _out, out):
        """
        Check parse results.
        """
        res = out.getvalue()
        msg = f"Test error; in: {_in!r}, out: {res!r}, expect: {_out!r}"
        assert res == _out, msg

    def test_encoding_detection_utf_content(self):
        html = b'<meta http-equiv="content-type" content="text/html; charset=UTF-8">'
        self.encoding_test(html, "utf-8")

    def test_encoding_detection_utf_charset(self):
        html = b'<meta charset="UTF-8">'
        self.encoding_test(html, "utf-8")

    def test_encoding_detection_iso_content(self):
        html = (
            b'<meta http-equiv="content-type" content="text/html; charset=ISO8859-1">'
        )
        self.encoding_test(html, "iso8859-1")

    def test_encoding_detection_iso_charset(self):
        html = b'<meta charset="ISO8859-1">'
        self.encoding_test(html, "iso8859-1")

    def test_encoding_detection_iso_bad_charset(self):
        html = b'<meta charset="hulla">'
        self.encoding_test(html, "ascii")

    def test_encoding_detection_iso_bad_content(self):
        html = b'<meta http-equiv="content-type" content="text/html; charset=blabla">'
        self.encoding_test(html, "ascii")

    def encoding_test(self, html, expected):
        # For encoding detection Beautiful Soup uses if available in order
        # of preference cchardet then chardet then charset-normalizer.
        # Results for html without a valid charset may differ
        # based on cchardet/chardet/charset-normalizer availability.
        soup = htmlsoup.make_soup(html)
        assert soup.original_encoding == expected



# Copyright (C) 2006-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test robots.txt parsing.
"""

import unittest
import linkcheck.robotparser2


class TestRobotsTxt(unittest.TestCase):
    """
    Test string formatting routines.
    """

    def setUp(self):
        """
        Initialize self.rp as a robots.txt parser.
        """
        self.rp = linkcheck.robotparser2.RobotFileParser(session=None)

    def test_robotstxt(self):
        lines = [
            "User-agent: *",
        ]
        self.rp.parse(lines)
        self.assertTrue(self.rp.mtime() > 0)
        self.assertEqual(str(self.rp), "\n".join(lines))

    def test_robotstxt2(self):
        lines = [
            "User-agent: *",
            "Disallow: /search",
        ]
        self.rp.parse(lines)
        self.assertEqual(str(self.rp), "\n".join(lines))

    def test_robotstxt3(self):
        lines = [
            "Disallow: /search",
            "",
            "Allow: /search",
            "",
            "Crawl-Delay: 5",
            "",
            "Blabla",
            "",
            "Bla: bla",
        ]
        self.rp.parse(lines)
        self.assertEqual(str(self.rp), "")

    def test_robotstxt4(self):
        lines = [
            "User-agent: Bla",
            "Disallow: /cgi-bin",
            "User-agent: *",
            "Disallow: /search",
        ]
        self.rp.parse(lines)
        lines.insert(2, "")
        self.assertEqual(str(self.rp), "\n".join(lines))

    def test_robotstxt5(self):
        lines = [
            "#one line comment",
            "User-agent: Bla",
            "Disallow: /cgi-bin # comment",
            "Allow: /search",
        ]
        lines2 = [
            "User-agent: Bla",
            "Disallow: /cgi-bin",
            "Allow: /search",
        ]
        self.rp.parse(lines)
        self.assertEqual(str(self.rp), "\n".join(lines2))

    def test_robotstxt6(self):
        lines = [
            "User-agent: Bla",
            "",
        ]
        self.rp.parse(lines)
        self.assertEqual(str(self.rp), "")

    def test_robotstxt7(self):
        lines = [
            "User-agent: Bla",
            "Allow: /",
            "",
            "User-agent: *",
            "Disallow: /",
        ]
        self.rp.parse(lines)
        self.assertEqual(str(self.rp), "\n".join(lines))
        self.assertTrue(self.rp.can_fetch("Bla", "/"))

    def test_crawldelay(self):
        lines = [
            "User-agent: Blubb",
            "Crawl-delay: 10",
            "",
            "User-agent: Hulla",
            "Crawl-delay: 5",
            "",
            "User-agent: *",
            "Crawl-delay: 1",
        ]
        self.rp.parse(lines)
        self.assertEqual(str(self.rp), "\n".join(lines))
        self.assertEqual(self.rp.get_crawldelay("Blubb"), 10)
        self.assertEqual(self.rp.get_crawldelay("Hulla"), 5)
        self.assertEqual(self.rp.get_crawldelay("Bulla"), 1)

    def test_crawldelay2(self):
        lines = [
            "User-agent: Blubb",
            "Crawl-delay: X",
        ]
        self.rp.parse(lines)
        del lines[1]
        self.assertEqual(str(self.rp), "\n".join(lines))

    def check_urls(self, good, bad, agent="test_robotparser"):
        for url in good:
            self.check_url(agent, url, True)
        for url in bad:
            self.check_url(agent, url, False)

    def check_url(self, agent, url, can_fetch):
        if isinstance(url, tuple):
            agent, url = url
        res = self.rp.can_fetch(agent, url)
        if can_fetch:
            self.assertTrue(res, "%s disallowed" % url)
        else:
            self.assertFalse(res, "%s allowed" % url)

    def test_access1(self):
        lines = [
            "User-agent: *",
            "Disallow: /cyberworld/map/ # This is an infinite virtual URL space",
            "Disallow: /tmp/ # these will soon disappear",
            "Disallow: /foo.html",
        ]
        lines2 = [
            "User-agent: *",
            "Disallow: /cyberworld/map/",
            "Disallow: /tmp/",
            "Disallow: /foo.html",
        ]
        self.rp.parse(lines)
        self.assertEqual(str(self.rp), "\n".join(lines2))
        good = ["/", "/test.html"]
        bad = ["/cyberworld/map/index.html", "/tmp/xxx", "/foo.html"]
        self.check_urls(good, bad)

    def test_access2(self):
        lines = [
            "# robots.txt for http://www.example.com/",
            "",
            "User-agent: *",
            "Disallow: /cyberworld/map/ # This is an infinite virtual URL space",
            "",
            "# Cybermapper knows where to go.",
            "User-agent: cybermapper",
            "Disallow:",
            "",
        ]
        lines2 = [
            "User-agent: cybermapper",
            "Allow: /",
            "",
            "User-agent: *",
            "Disallow: /cyberworld/map/",
        ]
        self.rp.parse(lines)
        self.assertEqual(str(self.rp), "\n".join(lines2))
        good = ["/", "/test.html", ("cybermapper", "/cyberworld/map/index.html")]
        bad = ["/cyberworld/map/index.html"]
        self.check_urls(good, bad)

    def test_access3(self):
        lines = [
            "# go away",
            "User-agent: *",
            "Disallow: /",
        ]
        lines2 = [
            "User-agent: *",
            "Disallow: /",
        ]
        self.rp.parse(lines)
        self.assertEqual(str(self.rp), "\n".join(lines2))
        good = []
        bad = ["/cyberworld/map/index.html", "/", "/tmp/"]
        self.check_urls(good, bad)

    def test_access4(self):
        lines = [
            "User-agent: figtree",
            "Disallow: /tmp",
            "Disallow: /a%3cd.html",
            "Disallow: /a%2fb.html",
            "Disallow: /%7ejoe/index.html",
        ]
        lines2 = [
            "User-agent: figtree",
            "Disallow: /tmp",
            "Disallow: /a%3Cd.html",
            "Disallow: /a/b.html",
            "Disallow: /~joe/index.html",
        ]
        self.rp.parse(lines)
        self.assertEqual(str(self.rp), "\n".join(lines2))
        good = []
        bad = [
            "/tmp",
            "/tmp.html",
            "/tmp/a.html",
            "/a%3cd.html",
            "/a%3Cd.html",
            "/a%2fb.html",
            "/~joe/index.html",
            "/a/b.html",
        ]
        self.check_urls(good, bad, "figtree")
        self.check_urls(good, bad, "FigTree/1.0 Robot libwww-perl/5.04")

    def test_access5(self):
        lines = [
            "User-agent: *",
            "Disallow: /tmp/",
            "Disallow: /a%3Cd.html",
            "Disallow: /a/b.html",
            "Disallow: /%7ejoe/index.html",
        ]
        lines2 = [
            "User-agent: *",
            "Disallow: /tmp/",
            "Disallow: /a%3Cd.html",
            "Disallow: /a/b.html",
            "Disallow: /~joe/index.html",
        ]
        self.rp.parse(lines)
        self.assertEqual(str(self.rp), "\n".join(lines2))
        good = ["/tmp"]  # XFAIL: '/a%2fb.html'
        bad = [
            "/tmp/",
            "/tmp/a.html",
            "/a%3cd.html",
            "/a%3Cd.html",
            "/a/b.html",
            "/%7Ejoe/index.html",
        ]
        self.check_urls(good, bad)

    def test_access6(self):
        lines = [
            "User-Agent: *",
            "Disallow: /.",
        ]
        self.rp.parse(lines)
        good = ["/foo.html"]
        bad = []  # Bug report says "/" should be denied, but that is not in the RFC
        self.check_urls(good, bad)

    def test_access7(self):
        lines = [
            "User-agent: Example",
            "Disallow: /example",
            "",
            "User-agent: *",
            "Disallow: /cgi-bin",
        ]
        self.rp.parse(lines)
        # test re.escape
        self.check_url("*", "/", True)
        # should match first agent
        self.check_url("", "/example", False)
        # test agent matching
        self.check_url("Example", "/example", False)
        self.check_url("Example/1.0", "/example", False)
        self.check_url("example", "/example", False)
        self.check_url("spam", "/cgi-bin", False)
        self.check_url("spam", "/cgi-bin/foo/bar", False)
        self.check_url("spam", "/cgi-bin?a=1", False)
        self.check_url("spam", "/", True)

    def test_sitemap(self):
        lines = [
            "Sitemap: bla",
        ]
        self.rp.parse(lines)
        self.assertTrue(len(self.rp.sitemap_urls) > 0)
        self.assertTrue(self.rp.sitemap_urls[0] == ("bla", 1))



# Copyright (C) 2020 Chris Mayo
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test login form functions.
"""

import unittest

from linkcheck.htmlutil import loginformsearch

login_form = """
<html>
<body>
<h1>Not a Login</h1>
<form action="/not_a_login">
<input name="some_field">
</form>
<h1>Login</h1>
<form action="/log_me_in">
<input name="User_Field">
<input name="Password_Field">
<input name="extra_field">
</form>
</body>
</html>
"""


class TestFormSearch(unittest.TestCase):
    """Test processing of a login form."""

    def test_search_form(self):
        form = loginformsearch.search_form(login_form, "User_Field", "Password_Field")
        self.assertIsNotNone(form)
        self.assertEqual(form.url, "/log_me_in")
        self.assertIn("User_Field", form.data)
        self.assertIn("Password_Field", form.data)

    def test_search_form_none(self):
        form = loginformsearch.search_form(login_form, "user_field", "password_field")
        self.assertIsNone(form)



# -*- coding: iso-8859-1 -*-
# Copyright (C) 2004-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test string formatting operations.
"""

import unittest
import os
import time
import linkcheck.strformat


class TestStrFormat(unittest.TestCase):
    """
    Test string formatting routines.
    """

    def test_unquote(self):
        # Test quote stripping.
        u = linkcheck.strformat.unquote
        self.assertEqual(u(""), "")
        self.assertEqual(u(None), None)
        self.assertEqual(u("'"), "'")
        self.assertEqual(u('"'), '"')
        self.assertEqual(u('""'), "")
        self.assertEqual(u("''"), "")
        self.assertEqual(u("'a'"), "a")
        self.assertEqual(u("'a\"'"), 'a"')
        self.assertEqual(u("'\"a'"), '"a')
        self.assertEqual(u('"a\'"'), "a'")
        self.assertEqual(u('"\'a"'), "'a")
        self.assertEqual(u("'a'", matching=True), "a")
        self.assertEqual(u('"a"', matching=True), "a")
        # even mis-matching quotes should be removed...
        self.assertEqual(u("'a\""), "a")
        self.assertEqual(u("\"a'"), "a")
        # ...but not when matching is True
        self.assertEqual(u("'a\"", matching=True), "'a\"")
        self.assertEqual(u("\"a'", matching=True), "\"a'")

    def test_wrap(self):
        # Test line wrapping.
        wrap = linkcheck.strformat.wrap
        s = "11%(sep)s22%(sep)s33%(sep)s44%(sep)s55" % {"sep": os.linesep}
        # testing width <= 0
        self.assertEqual(wrap(s, -1), s)
        self.assertEqual(wrap(s, 0), s)
        n = len(os.linesep)
        gap = " "
        s2 = "11%(gap)s22%(sep)s33%(gap)s44%(sep)s55" % {"sep": os.linesep, "gap": gap}
        # splitting lines
        self.assertEqual(wrap(s2, 2), s)
        # combining lines
        self.assertEqual(wrap(s, 4 + n), s2)
        # misc
        self.assertEqual(wrap(s, -1), s)
        self.assertEqual(wrap(s, 0), s)
        self.assertEqual(wrap(None, 10), None)
        self.assertFalse(linkcheck.strformat.get_paragraphs(None))

    def test_strsize(self):
        # Test byte size strings.
        self.assertRaises(ValueError, linkcheck.strformat.strsize, -1)
        self.assertEqual(linkcheck.strformat.strsize(0), "0B")
        self.assertEqual(linkcheck.strformat.strsize(1), "1B")
        self.assertEqual(linkcheck.strformat.strsize(2), "2B")
        self.assertEqual(linkcheck.strformat.strsize(1023, grouping=False), "1023B")
        self.assertEqual(linkcheck.strformat.strsize(1024), "1KB")
        self.assertEqual(linkcheck.strformat.strsize(1024 * 25), "25.00KB")
        self.assertEqual(linkcheck.strformat.strsize(1024 * 1024), "1.00MB")
        self.assertEqual(linkcheck.strformat.strsize(1024 * 1024 * 11), "11.0MB")
        self.assertEqual(linkcheck.strformat.strsize(1024 * 1024 * 1024), "1.00GB")
        self.assertEqual(linkcheck.strformat.strsize(1024 * 1024 * 1024 * 14), "14.0GB")

    def test_indent(self):
        s = "bla"
        self.assertEqual(linkcheck.strformat.indent(s, ""), s)
        self.assertEqual(linkcheck.strformat.indent(s, " "), " " + s)

    def test_stripurl(self):
        self.assertEqual(linkcheck.strformat.stripurl("a\tb"), "a\tb")
        self.assertEqual(linkcheck.strformat.stripurl(" a\t b"), "a\t b")
        self.assertEqual(linkcheck.strformat.stripurl(" ab\t\ra\nb"), "ab")
        self.assertEqual(linkcheck.strformat.stripurl(None), None)
        self.assertEqual(linkcheck.strformat.stripurl(""), "")

    def test_limit(self):
        self.assertEqual(linkcheck.strformat.limit("", 0), "")
        self.assertEqual(linkcheck.strformat.limit("a", 0), "")
        self.assertEqual(linkcheck.strformat.limit("1", 1), "1")
        self.assertEqual(linkcheck.strformat.limit("11", 1), "1...")

    def test_strtime(self):
        zone = linkcheck.strformat.strtimezone()
        t = linkcheck.strformat.strtime(0, func=time.gmtime)
        self.assertEqual(t, "1970-01-01 00:00:00" + zone)

    def test_duration_long(self):
        def duration(s):
            return linkcheck.strformat.strduration_long(s, do_translate=False)
        self.assertEqual(duration(-0.5), "-0.50 seconds")
        self.assertEqual(duration(0), "0.00 seconds")
        self.assertEqual(duration(0.9), "0.90 seconds")
        self.assertEqual(duration(1), "1 second")
        self.assertEqual(duration(2), "2 seconds")
        self.assertEqual(duration(60), "1 minute")
        self.assertEqual(duration(120), "2 minutes")
        self.assertEqual(duration(60 * 60), "1 hour")
        self.assertEqual(duration(60 * 60 * 24), "1 day")
        self.assertEqual(duration(60 * 60 * 24 * 365), "1 year")
        self.assertEqual(
            duration(60 * 60 * 24 * 365 + 60 * 60 * 24 + 2), "1 year, 1 day"
        )

    def test_ascii_safe(self):
        ascii_safe = linkcheck.strformat.ascii_safe
        self.assertEqual(ascii_safe("a"), "a")
        self.assertEqual(ascii_safe("ä"), "")

    def test_strip_control_chars(self):
        strip = linkcheck.strformat.strip_control_chars
        self.assertEqual(strip(""), "")
        self.assertEqual(strip("a"), "a")
        self.assertEqual(strip("ä"), "ä")
        self.assertEqual(strip("\x01"), "")



# Copyright (C) 2009-2010 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os
from linkcheck.logger.csvlog import CSVLogger

from .. import TestBase


class TestCsvLogger(TestBase):
    def test_parts(self):
        args = dict(
            filename=os.path.join(os.path.dirname(__file__), "testlog.csv"),
            parts=["realurl"],
            fileoutput=1,
            separator=";",
            quotechar='"',
        )
        logger = CSVLogger(**args)
        try:
            logger.start_output()
        finally:
            logger.end_output()
            os.remove(args["filename"])



# Copyright (C) 2005 Joe Wreschnig
# Copyright (C) 2005-2010 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test gettext .po files.
"""

import unittest
import subprocess
import glob
from tests import need_msgfmt, need_posix


pofiles = None


def get_pofiles():
    """Find all .po files in this source."""
    global pofiles
    if pofiles is None:
        pofiles = []
        pofiles.extend(glob.glob("po/*.po"))
        pofiles.extend(glob.glob("doc/i18n/locales/**/*.po", recursive=True))
    if not pofiles:
        raise Exception("No .po files found")
    return pofiles


class TestPo(unittest.TestCase):
    """Test .po file syntax."""

    @need_posix
    @need_msgfmt
    def test_pos(self):
        """Test .po files syntax."""
        for f in get_pofiles():
            cp = subprocess.run(["msgfmt", "-c", "-o", "-", f], capture_output=True)
            self.assertEqual(
                cp.returncode, 0,
                msg=f"PO-file syntax error in {f!r}: {str(cp.stderr, 'utf-8')}")


class TestGTranslator(unittest.TestCase):
    """GTranslator displays a middot · for a space. Unfortunately, it
    gets copied with copy-and-paste, what a shame."""

    def test_gtranslator(self):
        """Test all pofiles for GTranslator brokenness."""
        for f in get_pofiles():
            with open(f, encoding="UTF-8") as fd:
                self.check_file(fd, f)

    def check_file(self, fd, f):
        """Test for GTranslator broken syntax."""
        for line in fd:
            if line.strip().startswith("#"):
                continue
            self.assertFalse(
                "·" in line,
                f"Broken GTranslator copy/paste in {f!r}:\n{line}",
            )



# Copyright (C) 2004-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test http checking.
"""
from . import LinkCheckTest
from . import TestLogger


class AllPartsLogger(TestLogger):
    logparts = [
        "cachekey",
        "realurl",
        "name",
        "base",
        "info",
        "warning",
        "result",
        "url",
        "line",
        "col",
        "size",
        "parent_url",
        "page",
        "content_type",
    ]


class TestAllParts(LinkCheckTest):
    """
    Test that all parts of logger are working properly.
    """

    logger = AllPartsLogger

    def test_all_parts_linenos(self):
        self.file_test("all_parts_linenos.html")



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Logging configuration
"""
import logging.config
import sys
from . import ansicolor

# application log areas
LOG_ROOT = "linkcheck"
LOG_CMDLINE = "linkcheck.cmdline"
LOG_CHECK = "linkcheck.check"
LOG_CACHE = "linkcheck.cache"
LOG_THREAD = "linkcheck.thread"
LOG_PLUGIN = "linkcheck.plugin"
lognames = {
    "cmdline": LOG_CMDLINE,
    "checking": LOG_CHECK,
    "cache": LOG_CACHE,
    "thread": LOG_THREAD,
    "plugin": LOG_PLUGIN,
    "all": LOG_ROOT,
}

lognamelist = ", ".join(repr(name) for name in lognames)

# logging configuration
configdict = {
    'version': 1,
    'loggers': {},
    'root': {'level': 'WARN'},
    'incremental': True,
}


def init_log_config(handler=None):
    """Set up the application logging (not to be confused with check loggers).
    """
    for applog in lognames.values():
        # propagate except for root app logger 'linkcheck'
        propagate = applog != LOG_ROOT
        configdict['loggers'][applog] = dict(level='INFO', propagate=propagate)

    logging.config.dictConfig(configdict)
    if handler is None:
        handler = ansicolor.ColoredStreamHandler(strm=sys.stderr)
    add_loghandler(handler)


def add_loghandler(handler):
    """Add log handler to root logger and LOG_ROOT and set formatting."""
    format = "%(levelname)s %(name)s %(asctime)s %(threadName)s %(message)s"
    handler.setFormatter(logging.Formatter(format))
    logging.getLogger(LOG_ROOT).addHandler(handler)
    logging.getLogger().addHandler(handler)


def remove_loghandler(handler):
    """Remove log handler from root logger and LOG_ROOT."""
    logging.getLogger(LOG_ROOT).removeHandler(handler)
    logging.getLogger().removeHandler(handler)


def reset_loglevel():
    """Reset log level to display only warnings and errors."""
    set_loglevel(['all'], logging.WARN)


def set_debug(loggers):
    """Set debugging log level."""
    set_loglevel(loggers, logging.DEBUG)
    # enable for httplib debugging (used by requests.packages.urllib3)
    # import httplib
    # httplib.HTTPConnection.debuglevel = 1


def set_loglevel(loggers, level):
    """Set logging levels for given loggers."""
    if not loggers:
        return
    if 'all' in loggers:
        loggers = lognames.keys()
    for key in loggers:
        logging.getLogger(lognames[key]).setLevel(level)



# Copyright (C) 2004-2012 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test cgi form routines.
"""
import urllib.parse
from io import BytesIO
from wsgiref.util import setup_testing_defaults
from linkcheck.lc_cgi import checkform, checklink, LCFormError, application
from linkcheck.strformat import limit

from . import TestBase


class TestWsgi(TestBase):
    """Test wsgi application."""

    def test_form_valid_url(self):
        # Check url validity.
        env = dict()
        form = dict(url="http://www.example.com/", level="1")
        checkform(form, env)

    def test_form_empty_url(self):
        # Check with empty url.
        env = dict()
        form = dict(url="", level="0")
        self.assertRaises(LCFormError, checkform, form, env)

    def test_form_default_url(self):
        # Check with default url.
        env = dict()
        form = dict(url="http://", level="0")
        self.assertRaises(LCFormError, checkform, form, env)

    def test_form_invalid_url(self):
        # Check url (in)validity.
        env = dict()
        form = dict(url="http://www.foo bar/", level="0")
        self.assertRaises(LCFormError, checkform, form, env)

    def test_checklink(self):
        form = dict(url="http://www.example.com/", level="0")
        checklink(form, {})

    def test_application(self):
        form = dict(url="http://www.example.com/", level="0")
        formdata = urllib.parse.urlencode(form)
        formdata = formdata.encode("ascii")
        environ = {"wsgi.input": BytesIO(formdata)}
        setup_testing_defaults(environ)
        test_response = b""
        test_headers = [None]
        test_status = [None]

        def start_response(status, headers):
            test_status[0] = status
            test_headers[0] = headers

        for str_data in application(environ, start_response):
            if not isinstance(str_data, bytes):
                err = "answer is not a byte string: %r" % limit(str_data, 30)
                self.assertTrue(False, err)
            test_response += str_data
        self.assertEqual(test_status[0], "200 OK")
        self.assertTrue(b"Generated by LinkChecker" in test_response)
        self.assertFalse(b"internal error" in test_response)
